import numpy as np
import logging
import time
import pysnptools.util as pstutil
from pysnptools.pstreader import PstReader
import warnings
import pysnptools.standardizer as stdizer
from bed_reader import open_bed


# !!why do the examples use ../tests/datasets instead of "examples"?
class SnpReader(PstReader):
    """A SnpReader is one of three things:

    * A class such as :class:`.Bed` for you to specify a file with data. For example,

    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snp_on_disk = Bed(bedfile,count_A1=False)
    >>> print(snp_on_disk) # prints the name of the file reader
    Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False)
    >>> snp_on_disk.sid_count # prints the number of SNPS (but doesn't read any SNP values)
    1015

    * A :class:`.SnpData` class that holds SNP data in memory, typically after reading it from disk:

    >>> snp_on_disk = Bed(bedfile,count_A1=False)
    >>> snpdata1 = snp_on_disk.read() #reads the SNP values
    >>> type(snpdata1.val).__name__ # The val property is an ndarray of SNP values
    'ndarray'
    >>> print(snpdata1) # prints the name of the file reader
    SnpData(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False))
    >>> snpdata1.iid_count #prints the number of iids (number of individuals) in this in-memory data
    300

    * A subset of any SnpReader, specified with "[ *iid_index* , *sid_index* ]", to read only some SNP values.
        It can also be used to re-order the values.

    >>> snp_on_disk = Bed(bedfile,count_A1=False)
    >>> subset_on_disk = snp_on_disk[[3,4],::2] # specification for a subset of the data on disk. No SNP values are read yet.
    >>> print(subset_on_disk.sid_count) # prints the number of sids in this subset (but still doesn't read any SNP values)
    508
    >>> print(subset_on_disk) #prints a specification of 'subset_on_disk'
    Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False)[[3,4],::2]
    >>> snpdata_subset = subset_on_disk.read() # efficiently reads the specified subset of values from the disk
    >>> print(snpdata_subset) # prints the name of the file reader
    SnpData(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False)[[3,4],::2])
    >>> print((int(snpdata_subset.val.shape[0]), int(snpdata_subset.val.shape[1]))) # The dimensions of the ndarray of SNP values
    (2, 508)

        The SnpReaders Classes

            ========================= =================== ====================== ================== ======================
            *Class*                   *Format*            *Random Access*        *Suffixes*         *Write* method?
            :class:`.SnpData`         in-memory floats    Yes                    *n/a*              *n/a*
            :class:`.Bed`             binary, 0,1,2       Yes (by sid)           .bed/.bim/.fam     Yes
            :class:`.Pheno`           text, floats        No                     .txt,.phe          Yes
            :class:`.Dat`             text, floats        No                     .dat/.map/.fam     Yes
            :class:`.Ped`             text, 0,1,2         No                     .ped/.map          Yes
            :class:`.Dense`           text, 0,1,2         No                     .dense.txt         Yes
            :class:`.SnpNpz`          binary, floats      No                     .snp.npz           Yes
            :class:`.SnpHdf5`         binary, floats      Yes (by sid or iid)    .snp.hdf5          Yes
            :class:`.SnpMemMap`       mem-mapped floats   Yes                    .snp.memmap        Yes
            :class:`.SnpGen`          generated values    Yes (by sid)           *n/a*              *n/a*
            ========================= =================== ====================== ================== ======================


        Methods & Properties:

            Every SnpReader, such as :class:`.Bed` and :class:`.SnpData`, has these properties: :attr:`iid`, :attr:`iid_count`, :attr:`sid`, :attr:`sid_count`,
            :attr:`pos` and these methods: :meth:`read`, :meth:`iid_to_index`, :meth:`sid_to_index`, :meth:`read_kernel`. See below for details.

            :class:`.SnpData` is a SnpReader so it supports the above properties and methods. In addition, it supports property :attr:`.SnpData.val`,
            method :meth:`.SnpData.standardize`, and equality testing.
            See below for details.

            Many of the classes, such as :class:`.Bed`, also provide a static :meth:`Bed.write` method for writing :class:`.SnpData` to disk.

            >>> # read from Pheno, write to Bed
            >>> from pysnptools.snpreader import Pheno, Bed
            >>> import pysnptools.util as pstutil
            >>> from pysnptools.util import example_file # Download and return local file name

            >>> bed_fn = example_file("pysnptools/examples/toydata.5chrom.*", "*.bed")
            >>> snpdata = Bed(bed_fn)[:,::2].read() # Read every-other SNP
            >>> pstutil.create_directory_if_necessary("tempdir/everyother.bed")
            >>> Bed.write("tempdir/everyother.bed",snpdata,count_A1=False)   # Write data in Bed format
            Bed('tempdir/everyother.bed',count_A1=False)


        iids and sids:

            Individuals are identified with an iid, which is a ndarray of two strings: a family ID and a individual ID.
            SNP locations are identified with an sid string in a ndarray. For example:

            >>> snp_on_disk = Bed(bedfile,count_A1=False)
            >>> print(snp_on_disk.iid[:3]) # print the first three iids
            [['POP1' '0']
             ['POP1' '12']
             ['POP1' '44']]
            >>> print(snp_on_disk.sid[:9]) # print the first nine sids
            ['1_12' '1_34' '1_10' '1_35' '1_28' '1_25' '1_36' '1_39' '1_4']
            >>> print(snp_on_disk.iid_to_index([['POP1','44'],['POP1','12']])) #Find the indexes for two iids.
            [2 1]

        Selecting and Reordering Individuals and SNPs

            You often don't want to read the SNP values for all iids and sids. You can use indexing to create a subsetting SnpReader that
            will read only the SNP values of interest.

            SnpReaders support the indexing formats supported by ndarray plus two generalizations. Here are examples of indexing with an array
            of indexes, with slicing, and with an array of Booleans.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> subset_snpreader_1 = snp_on_disk[[3,-1],:] #index with an array of indexes. Negatives count from end.
                >>> print((subset_snpreader_1.iid_count, subset_snpreader_1.sid_count))
                (2, 1015)
                >>> snpdata1 = subset_snpreader_1.read() # read just the two rows of interest from the disk
                >>> subset_snpreader_2 = snp_on_disk[:,:0:-2] #index with a slice
                >>> print((subset_snpreader_2.iid_count, subset_snpreader_2.sid_count))
                (300, 507)
                >>> boolindexes = [s.startswith('23_') for s in snp_on_disk.sid] # create a Boolean index of sids that start '23_'
                >>> subset_snpreader_3 = snp_on_disk[:,boolindexes] #index with array of Booleans
                >>> print((subset_snpreader_3.iid_count, subset_snpreader_3.sid_count))
                (300, 24)

            The first generalization over what ndarray offers is full indexing on both the iid dimension and the sid dimension, in other words,
            full multidimensional indexing. For example,

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> subset_snpreader_4 = snp_on_disk[[3,4],:0:-2] # index on two dimensions at once
                >>> print((subset_snpreader_4.iid_count, subset_snpreader_4.sid_count))
                (2, 507)

            The second generalization is indexing on a single integer index.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> subset_snpreader_5 = snp_on_disk[5,:] #index with single integer
                >>> print((subset_snpreader_5.iid_count, subset_snpreader_5.sid_count))
                (1, 1015)

            Indexing is also useful when you have SNP values in memory via a :class:`SnpData` index and want to copy a subset of those values.
            While you could instead index directly on the :attr:`SnpData.val` ndarray, by indexing on the :class:`SnpData` instance you
            also get iid and cid information.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> snpdata1 = snp_on_disk.read() # read all SNP values into memory
                >>> print(snpdata1.sid[:9]) # print the first 9 sids
                ['1_12' '1_34' '1_10' '1_35' '1_28' '1_25' '1_36' '1_39' '1_4']
                >>> snpdata_subset = snpdata1[:,::2].read(view_ok=True,order='A') # create a copy or view with every other sid
                >>> print(snpdata_subset.sid[:9])# print the first 9 sids in the subset
                ['1_12' '1_10' '1_28' '1_36' '1_4' '1_11' '1_32' '1_9' '1_17']


            You can apply indexing on top of indexing to specify subsets of subsets of data to read. In this example,
            only the SNP values for every 16th sid is actually read from the disk.

                >>> # These are just SnpReaders, nothing is read from disk yet
                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> half_snpreader = snp_on_disk[:,::2] # a reader for half the sids
                >>> quarter_snpreader = half_snpreader[:,::2] # a reader for half of half the sids
                >>> sixteenth_snpreader = quarter_snpreader[:,::2][:,::2] # a reader for half of half of half of half the sids
                >>> print(sixteenth_snpreader) #Print the specification of this reader
                Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False)[:,::2][:,::2][:,::2][:,::2]
                >>> # Now we read from disk. Only values for one sid in every 16 will be read.
                >>> snpdata_sixteenth = sixteenth_snpreader.read()
                >>> print(snpdata_sixteenth.val[0,3])
                2.0

        When Data is Read:

            SNP data can be enormous so we generally avoid reading it to the degree practical. Specifically,

            * Constructing and printing a SnpReader causes no file reading. For example, these commands read no data:

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> print(snp_on_disk) # Print the Bed SnpReader specification. No data is read.
                Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False)
                >>> subset_on_disk = snp_on_disk[[3,4],::2] # Construct a subsetting SnpReader. No data is read.
                >>> print(subset_on_disk) # print the subset SnpReader. No data is read.
                Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False)[[3,4],::2]

            * Properties and methods related to the iids and sids (to the degree practical) read just some iid and sid data from the disk,
              not SNP value data. Moreover, the iid and sid data is read from file only once. Consider these commands:

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> print(snp_on_disk.sid[:9]) # without reading any SNP values data from disk, read the sid and iid data from disk, cache it, and then print the first nine sids.
                ['1_12' '1_34' '1_10' '1_35' '1_28' '1_25' '1_36' '1_39' '1_4']
                >>> print(snp_on_disk.sid_to_index(['1_10','1_13'])) #use the cached sid information to find the indexes of '1_10' and '1_13'. (No data is read from disk.)
                [2 9]

            * The only methods that read SNP values from file are :meth:`read` and :meth:`read_kernel` (to the degree practical). For example:

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> snpdata1 = snp_on_disk.read() #read all the SNP values from disk, creating a new SnpData instance that keeps these values in memory
                >>> print(snpdata1.val[0,2])# print the SNP value for the iid with index 0 and the sid with index 2. (No data is read from disk.)
                1.0

            * If you request the values for only a subset of the iids or sids, (to the degree practical) only that subset will be read from disk.
              for example:

                >>> subset_on_disk = Bed(bedfile,count_A1=False)[[3,4],::2] # Construct a subsetting SnpReader. No data is read.
                >>> snpdata_subset = subset_on_disk.read() # from disk, read the SNP values for the iids with index 3 and 4 AND sids with even numbered indexes.
                >>> print(snpdata_subset.val[0,2]) # print the SNP value with subset iid index 0 and sid index 2 (corresponding to iid index 3 and sid index 4 in the full data). No data is read from disk.
                2.0

        When Data is Re-Read and Copied:

            Every time you call a SnpReader's :meth:`read` method, the SnpReader re-reads the SNP value data and returns a new in-memory :class:`.SnpData`
            (with :attr:`.SnpData.val` property containing a new ndarray of the SNP values). Likewise, when you call the :meth:`kernel` method, the SnpReader re-reads
            the data and returns a new kernel ndarray.

            Here is an example of what not to do, because it causes all the SNP value data to be read twice.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> # The following is not recommended because it inefficiently reads all the SNP values twice.
                >>> print(snp_on_disk.read().val[0,2]) # read all values into a new SnpData, print a SNP value
                1.0
                >>> print(snp_on_disk.read().val[0,3]) # read all values (again) into a second new SnpData, print a SNP value
                2.0

            Here are two efficient alternatives. First, if all SNP values can all fit in memory, read them once into a :class:`SnpData` and then
            access that :class:`SnpData` multiple times.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> snpdata1 = snp_on_disk.read() # read all values into a new SnpData
                >>> print(snpdata1.val[0,2]) # print a SNP value from snpdata1's in-memory ndarray
                1.0
                >>> print(snpdata1.val[0,3]) # print another SNP value from snpdata1's in-memory ndarray.
                2.0

            Second, if the SNP value data is too large to fit in memory, use subsetting to read only the SNP values of interest from disk.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> print(snp_on_disk[0,2].read().val[0,0]) #Define the subset of data and read only that subset from disk.
                1.0
                >>> print(snp_on_disk[0,3].read().val[0,0]) #Define a second subset of data and read only that subset from disk.
                2.0

            Because the in-memory :class:`.SnpData` class is a kind of SnpReader, you may read from it, too.
            Doing so create a new :class:`.SnpData` instance containing a copy of the SNP values in a new ndarray.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> snpdata1 = snp_on_disk.read() # read all SNP values from disk into a new SnpData
                >>> print(snpdata1.val is snpdata1.val) # Do the in-memory SNP values use the same memory as themselves? Yes
                True
                >>> snpdata2 = snpdata1.read() # copy all the SNP values into a new ndarray in a new SnpData
                >>> print(snpdata2.val is snpdata1.val) # Do the two ndarrays of in-memory SNP values use the same memory?
                False


        Avoiding Unwanted ndarray Allocations

            You may want a subset of SNPs values from an in-memory :class:`SnpData` and you may know that this subset and the original :class:`SnpData`
            can safely share the memory of the ndarray of SNP values. For this case, the :meth:`read` has optional parameters called view_ok and order. If you override
            the defaults of "view_ok=False,order='F'" with "view_ok=True,order='A', the :meth:`read` will, if practical, return a new
            :class:`SnpData` with a ndarray shares memory with the original ndarray.
            Use these parameters with care because any change to either ndarray (for example, via :meth:`.SnpData.standardize`) will effect
            the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
            share memory and so it may ignore your suggestion and allocate a new ndarray anyway.

    snp_on_disk = Bed(bedfile,count_A1=False) # Construct a Bed SnpReader. No data is read.
                >>> snpdata1 = snp_on_disk.read() # read all data from disk into a SnpData with a new ndarray
                >>> column01 = snpdata1[:,0:1].read(view_ok=True,order='A') #create SnpData with the data from just the first two SNPs. Sharing memory is OK. The memory may be laid out in any order (that is sid-major and iid-major are both OK).
                >>> import numpy as np
                >>> #print np.may_share_memory(snpdata1.val, column01.val) # Do the two ndarray's share memory? They could (but currently they won't)
                >>> column201 = snpdata1[:,[2,0,1]].read(view_ok=True,order='A') #create SnpData with the data from three SNPs, permuted. Sharing memory is OK.
                >>> print(np.may_share_memory(snpdata1.val, column201.val)) # Do the two ndarray's share memory? No, ndarray decided that this indexing was too complex for sharing.
                False

        The :meth:`read` Method

            By default the :meth:`read` returns a ndarray of numpy.float64 laid out in memory in F-contiguous order (iid-index varies the fastest). You may, instead,
            ask for numpy.float32 or for C-contiguous order or any order. See :meth:`read` for details.

        The :meth:`.SnpData.standardize` Method
            The :meth:`.SnpData.standardize` method, available only on :class:`.SnpData`, does in-place standardization of the in-memory
            SNP data. By default, it applies 'Unit' standardization, that is: the values for each SNP will have mean zero and standard deviation 1.0.
            NaN values are then filled with zeros (consequently, if there are NaN values, the final standard deviation will not be zero).
            Note that, for efficiently, this method works in-place, actually changing values in the ndarray. Although it works in place, for convenience
            it also returns itself. See :meth:`.SnpData.standardize` for options and details.

                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> snpdata1 = snp_on_disk.read() # read all SNP values into memory
                >>> print(snpdata1) # Prints the specification for this SnpData
                SnpData(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False))
                >>> print(snpdata1.val[0,0])
                2.0
                >>> snpdata1.standardize() # standardize changes the values in snpdata1.val and changes the specification.
                SnpData(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False),Unit())
                >>> print('{0:.6f}'.format(snpdata1.val[0,0]))
                0.229416
                >>> snpdata2 = snp_on_disk.read().standardize() # Read and standardize in one expression with only one ndarray allocated.
                >>> print('{0:.6f}'.format(snpdata2.val[0,0]))
                0.229416

        The :meth:`read_kernel` Method

            The :meth:`read_kernel` method, available on any SnpReader, returns a :class:`KernelData`. The :attr:`val` property of the :class:`KernelData` is
            an ndarray of the (possibility standardized) SNP values multiplied with their transposed selves. When applied to an read-from-disk SnpReader, such as :class:`.Bed`,
            the method can save memory by reading (and standardizing) the data in blocks. See :meth:`read_kernel` for details.


                >>> from pysnptools.standardizer import Unit
                >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify some data on disk in Bed format
                >>> kerneldata1 = snp_on_disk.read_kernel(Unit()) #Create an in-memory kernel from the snp data on disk.
                >>> print('{0:.6f}'.format(kerneldata1.val[0,0]))
                901.421836
                >>> kerneldata2 = snp_on_disk.read_kernel(Unit(),block_size=10) #Create an in-memory kernel from the snp data on disk, but only read 10 SNPS at a time from the disk.
                >>> print('{0:.6f}'.format(kerneldata2.val[0,0]))
                901.421836

        Details of Methods & Properties:
    """

    def __init__(self, *args, **kwargs):
        super(SnpReader, self).__init__(*args, **kwargs)

    @property
    def iid(self):
        """A ndarray of the iids. Each iid is a ndarray of two strings (a family ID and a individual ID) that identifies an individual.

        :rtype: ndarray of strings with shape [:attr:`.iid_count`,2]

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False)
        >>> print(snp_on_disk.iid[:3]) # print the first three iids
        [['POP1' '0']
         ['POP1' '12']
         ['POP1' '44']]
        """
        return self.row

    @property
    def iid_count(self):
        """number of iids

        :rtype: integer

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.
        """
        return self.row_count

    @property
    def sid(self):
        """A ndarray of the sids. Each sid is a string that identifies a SNP.

        :rtype: ndarray (length :attr:`.sid_count`) of strings

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False)
        >>> print(snp_on_disk.sid[:9]) # print the first nine sids
        ['1_12' '1_34' '1_10' '1_35' '1_28' '1_25' '1_36' '1_39' '1_4']

        """
        return self.col

    @property
    def sid_count(self):
        """number of sids

        :rtype: integer

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        """
        return self.col_count

    # !!document that chr must not be X,Y,M only numbers (as per the PLINK BED format)
    # !!Also what about telling the ref and alt allele? Also, what about tri and quad alleles, etc?
    @property
    def pos(self):
        """A ndarray of the position information for each sid. Each element is a ndarray of three numpy.numbers (chromosome, genetic distance, basepair distance).

        :rtype: ndarray of float64 with shape [:attr:`.sid_count`, 3]

        This property (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False)
        >>> print(snp_on_disk.pos[:3,]) # print position information for the first three sids: #The '...' is for possible space char
        [[1.         0.00800801        nan]
         [1.         0.023023   1.        ]
         [1.         0.0700701  4.        ]]
        """
        return self.col_property

    @property
    def row_property(self):
        """Defined as a zero-width array for compatibility with :class:`PstReader`, but not used."""
        if not hasattr(self, "_row_property"):
            self._row_property = np.empty((self.row_count, 0))
        return self._row_property

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        raise NotImplementedError

    # !!check that views always return contiguous memory by default
    def read(
        self,
        order="F",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        num_threads=None,
        _require_float32_64=True,
    ):
        """Reads the SNP values and returns a :class:`.SnpData` (with :attr:`.SnpData.val` property containing a new ndarray of the SNP values).

        :param order: {'F' (default), 'C', 'A'}, optional -- Specify the order of the ndarray. If order is 'F' (default),
            then the array will be in F-contiguous order (iid-index varies the fastest).
            If order is 'C', then the returned array will be in C-contiguous order (sid-index varies the fastest).
            If order is 'A', then the :attr:`.SnpData.val`
            ndarray may be in any order (either C-, Fortran-contiguous).
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`SnpData.val` ndarray.
             (For :class:`Bed`, only, it can also be numpy.int8. Hidden option _require_float32_64 must be set
             to False. See :class:`Bed` for an example.)
        :type dtype: data-type.

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool

        :param view_ok: optional -- If False (default), allocates new memory for the :attr:`SnpData.val`'s ndarray. If True,
            if practical and reading from a :class:`SnpData`, will return a new
            :class:`SnpData` with a ndarray shares memory with the original :class:`SnpData`.
            Typically, you'll also wish to use "order='A'" to increase the chance that sharing will be possible.
            Use these parameters with care because any change to either ndarray (for example, via :meth:`.SnpData.standardize`) will effect
            the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
            share memory and so it may ignore your suggestion and allocate a new ndarray anyway.
        :type view_ok: bool

        :param num_threads: optional -- The number of threads with which to read data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int

        :rtype: :class:`.SnpData`

        Calling the method again causes the SNP values to be re-read and creates a new in-memory :class:`.SnpData` with a new ndarray of SNP values.

        If you request the values for only a subset of the sids or iids, (to the degree practical) only that subset will be read from disk.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile, count_A1=False) # Specify SNP data on disk
        >>> snpdata1 = snp_on_disk.read() # Read all the SNP data returning a SnpData instance
        >>> print(type(snpdata1.val).__name__) # The SnpData instance contains a ndarray of the data.
        ndarray
        >>> subset_snpdata = snp_on_disk[:,::2].read() # From the disk, read SNP values for every other sid
        >>> print(subset_snpdata.val[0,0]) # Print the first SNP value in the subset
        2.0
        >>> subsub_snpdata = subset_snpdata[:10,:].read(order='A',view_ok=True) # Create an in-memory subset of the subset with SNP values for the first ten iids. Share memory if practical.
        >>> import numpy as np
        >>> # print np.may_share_memory(subset_snpdata.val, subsub_snpdata.val) # Do the two ndarray's share memory? They could. Currently they won't.
        """
        dtype = np.dtype(dtype)
        val = self._read(
            None, None, order, dtype, force_python_only, view_ok, num_threads
        )
        from pysnptools.snpreader import SnpData

        ret = SnpData(
            self.iid,
            self.sid,
            val,
            pos=self.pos,
            name=str(self),
            _require_float32_64=_require_float32_64,
        )
        return ret

    def iid_to_index(self, list):
        """Takes a list of iids and returns a list of index numbers

        :param list: list of iids
        :type order: list of list of strings

        :rtype: ndarray of int

        This method (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify SNP data on disk
        >>> print(snp_on_disk.iid_to_index([['POP1','44'],['POP1','12']])) #Find the indexes for two iids.
        [2 1]
        """
        return self.row_to_index(list)

    def sid_to_index(self, list):
        """Takes a list of sids and returns a list of index numbers

        :param list: list of sids
        :type list: list of strings

        :rtype: ndarray of int

        This method (to the degree practical) reads only iid and sid data from the disk, not SNP value data. Moreover, the iid and sid data is read from file only once.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify SNP data on disk
        >>> print(snp_on_disk.sid_to_index(['1_10','1_13'])) #Find the indexes for two sids.
        [2 9]
        """
        return self.col_to_index(list)

    def __getitem__(self, iid_indexer_and_snp_indexer):
        from pysnptools.snpreader._subset import _SnpSubset

        iid_indexer, snp_indexer = iid_indexer_and_snp_indexer
        return _SnpSubset(self, iid_indexer, snp_indexer)

    def read_kernel(
        self,
        standardizer=None,
        block_size=None,
        order="A",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        num_threads=None,
    ):
        """Returns a :class:`KernelData` such that the :meth:`KernelData.val` property will be a ndarray of the standardized SNP values multiplied with their transposed selves.

        :param standardizer: -- (required) Specify standardization to be applied before the matrix multiply. Any :class:`.Standardizer` may be used. Some choices include :class:`Standardizer.Identity`
            (do nothing), :class:`.Unit` (make values for each SNP have mean zero and standard deviation 1.0) and :class:`Beta`.
        :type standardizer: :class:`.Standardizer`

        :param block_size: optional -- Default of None (meaning to load all). Suggested number of sids to read into memory at a time.
        :type block_size: int or None

        :rtype: class:`KernelData`

        Calling the method again causes the SNP values to be re-read and allocates a new class:`KernelData`.

        When applied to an read-from-disk SnpReader, such as :class:`.Bed`, the method can save memory by reading (and standardizing) the data in blocks.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.standardizer import Unit
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify SNP data on disk
        >>> kerneldata1 = snp_on_disk.read_kernel(Unit())
        >>> print((int(kerneldata1.iid_count), '{0:.6f}'.format(kerneldata1.val[0,0])))
        (300, '901.421836')
        """
        assert standardizer is not None, "'standardizer' must be provided"
        dtype = np.dtype(dtype)

        from pysnptools.kernelreader import SnpKernel

        snpkernel = SnpKernel(self, standardizer=standardizer, block_size=block_size)
        kerneldata = snpkernel.read(
            order, dtype, force_python_only, view_ok, num_threads
        )
        return kerneldata

    def kernel(
        self,
        standardizer,
        allowlowrank=False,
        block_size=10000,
        blocksize=None,
        num_threads=None,
    ):
        """.. Warning:: Deprecated. Use :meth:`read_kernel` instead.

        Returns a ndarray of size iid_count x iid_count. The returned array has the value of the standardized SNP values multiplied with their transposed selves.

        :param standardizer: -- Specify standardization to be applied before the matrix multiply. Any :class:`.Standardizer` may be used. Some choices include :class:`Standardizer.Identity`
            (do nothing), :class:`.Unit` (make values for each SNP have mean zero and standard deviation 1.0), :class:`Beta`.
        :type standardizer: :class:`.Standardizer`

        :param block_size: optional -- Default of 10000. None means to load all. Suggested number of sids to read into memory at a time.
        :type block_size: int or None

        :param num_threads: optional -- The number of threads with which to standardize data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int


        :rtype: ndarray of size :attr:`.iid_count` x :attr:`.iid_count`

        Calling the method again causes the SNP values to be re-read and allocates a new ndarray.

        When applied to an read-from-disk SnpReader, such as :class:`.Bed`, the method can save memory by reading (and standardizing) the data in blocks.

        If you request the values for only a subset of the sids or iids, (to the degree practical) only that subset will be read from disk.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.standardizer import Unit
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bedfile,count_A1=False) # Specify SNP data on disk
        >>> kernel = snp_on_disk.kernel(Unit())
        >>> print(((int(kernel.shape[0]),int(kernel.shape[1])), '{0:.6f}'.format(kernel[0,0])))
        ((300, 300), '901.421836')
        """  # print "entering kernel with {0},{1},{2}".format(self, standardizer, blocksize)
        warnings.warn(
            ".kernel(...) is deprecated. Use '.read_kernel(...).val", DeprecationWarning
        )
        if blocksize is not None:
            warnings.warn(
                ".kernel(...blocksize...) is deprecated. Use '.kernel(...block_size=...)",
                DeprecationWarning,
            )
            block_size = blocksize
        return self._read_kernel(
            standardizer, block_size=block_size, num_threads=num_threads
        )

    @staticmethod
    def _as_snpdata(
        snpreader, standardizer, force_python_only, order, dtype, num_threads
    ):
        """
        Like 'read' except (1) won't read if already a snpdata and (2) returns the standardizer
        """
        dtype = np.dtype(dtype)

        if (
            hasattr(snpreader, "val")
            and snpreader.val.dtype == dtype
            and isinstance(standardizer, stdizer.Identity)
            and (
                order == "A"
                or (order == "C" and snpreader.val.flags["C_CONTIGUOUS"])
                or (order == "F" and snpreader.val.flags["F_CONTIGUOUS"])
            )
        ):
            return snpreader, stdizer.Identity()
        else:
            return snpreader.read(order=order, dtype=dtype).standardize(
                standardizer,
                return_trained=True,
                force_python_only=force_python_only,
                num_threads=num_threads,
            )

    def _read_kernel(
        self,
        standardizer,
        block_size=None,
        order="A",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        return_trained=False,
        num_threads=None,
    ):
        """
        Will respect the cupy environment variable.
        """
        dtype = np.dtype(dtype)
        # Do all-at-once (not in blocks) if 1. No block size is given or 2. The #ofSNPs < Min(block_size,iid_count)
        if block_size is None or (
            self.sid_count <= block_size or self.sid_count <= self.iid_count
        ):
            train_data, trained_standardizer = SnpReader._as_snpdata(
                self,
                standardizer=standardizer,
                dtype=dtype,
                order="A",
                force_python_only=force_python_only,
                num_threads=num_threads,
            )
            kernel = train_data._read_kernel(
                stdizer.Identity(),
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=False,
                num_threads=num_threads,
            )
            if return_trained:
                return kernel, trained_standardizer
            else:
                return kernel

        else:  # Do in blocks
            xp = pstutil.array_module()
            # Set the default order to 'C' because with kernels any order is fine and the Python .dot method likes 'C' best.
            if order == "A":
                order = "C"
            t0 = time.time()
            K = xp.zeros([self.iid_count, self.iid_count], dtype=dtype, order=order)
            trained_standardizer_list = []
            logging.info(
                f"reading {self.sid_count:,} SNPs in blocks of {block_size:,} and adding up kernels (for {self.iid_count:,} individuals) with {xp.__name__}."
            )

            ct = 0
            ts = time.time()
            diff_last = 0

            for start in range(0, self.sid_count, block_size):
                ct += block_size
                train_data, trained_standardizer = SnpReader._as_snpdata(
                    self[:, start : start + block_size],
                    standardizer=standardizer,
                    dtype=dtype,
                    order="A",
                    force_python_only=force_python_only,
                    num_threads=num_threads,
                )
                trained_standardizer_list.append(trained_standardizer)
                K += train_data._read_kernel(
                    stdizer.Identity(),
                    block_size=None,
                    order=order,
                    dtype=dtype,
                    force_python_only=force_python_only,
                    view_ok=False,
                    num_threads=num_threads,
                )
                if ct % block_size == 0:
                    diff = time.time() - ts
                    if diff > 1 and diff - diff_last > 5:
                        logging.info(f"read {ct:,} SNPs in {diff:.2f} seconds")
                        diff_last = diff

            t1 = time.time()
            logging.info("%.2f seconds elapsed" % (t1 - t0))

            if return_trained:
                return K, standardizer._merge_trained(
                    trained_standardizer_list
                )  # turns this into a single standardizer
            else:
                return K

    def copyinputs(self, copier):
        raise NotImplementedError

    def _assert_iid_sid_pos(self, check_val):
        if check_val:
            assert len(self._val.shape) == 2, "val should have two dimensions"
            assert self._val.shape == (
                len(self._row),
                len(self._col),
            ), "val shape should match that of iid_row x sid_row"

        assert (
            self._row.dtype.type is np.str_
            and len(self._row.shape) == 2
            and self._row.shape[1] == 2
        ), "iid should be dtype str, have two dimensions, and the second dimension should be size 2"
        assert (
            self._col.dtype.type is np.str_ and len(self._col.shape) == 1
        ), "sid should be of dtype of str and one dimensional"

    @staticmethod
    def _name_of_other_file(filename, remove_suffix, add_suffix):
        filename = str(filename)
        if filename.lower().endswith(remove_suffix.lower()):
            filename = filename[: -1 - len(remove_suffix)]
        return filename + "." + add_suffix

    @staticmethod
    def _write_fam(snpdata, basefilename, remove_suffix, add_suffix="fam"):
        famfile = SnpReader._name_of_other_file(basefilename, remove_suffix, add_suffix)

        with open(famfile, "w") as fam_filepointer:
            for iid_row in snpdata.iid:
                fam_filepointer.write(
                    "{0} {1} 0 0 0 0\n".format(iid_row[0], iid_row[1])
                )

    @staticmethod
    def _write_map_or_bim(snpdata, basefilename, remove_suffix, add_suffix):
        mapfile = SnpReader._name_of_other_file(basefilename, remove_suffix, add_suffix)
        with open(mapfile, "w") as map_filepointer:
            for sid_index, sid in enumerate(snpdata.sid):
                posrow = snpdata.pos[sid_index].copy()
                posrow[
                    posrow != posrow
                ] = 0  # Fill missing with zero as per the PLINK standard
                map_filepointer.write(
                    "%r\t%s\t%r\t%d\tA\tC\n" % (posrow[0], sid, posrow[1], posrow[2])
                )

    @staticmethod
    def _read_fam(basefilename, remove_suffix, add_suffix="fam"):
        famfile = SnpReader._name_of_other_file(basefilename, remove_suffix, add_suffix)

        properties = {
            "father": None,
            "mother": None,
            "sex": None,
            "pheno": None,
        }

        logging.info("Loading {0} file {1}".format(add_suffix, famfile))
        with open_bed(
            basefilename,
            fam_filepath=famfile,
            properties=properties,
            skip_format_check=True,
        ) as bed:
            iid = np.array([bed.fid, bed.iid]).T
        return iid

    @staticmethod
    def _read_map_or_bim(basefilename, remove_suffix, add_suffix):
        bimfile = SnpReader._name_of_other_file(basefilename, remove_suffix, add_suffix)

        properties = {
            "allele_1": None,
            "allele_2": None,
        }

        logging.info("Loading {0} file {1}".format(add_suffix, bimfile))
        with open_bed(
            basefilename,
            bim_filepath=bimfile,
            properties=properties,
            skip_format_check=True,
        ) as bed:
            sid = bed.sid
            pos = np.array(
                [
                    bed.chromosome.astype("float"),
                    bed.cm_position,
                    bed.bp_position,
                ]
            ).T  # LATER could copy in batches to use less memory
            pos[pos == 0] = np.nan

        return sid, pos

    @property
    def val_shape(self):
        """
        Tells the shape of value for a given individual and SNP. For SnpReaders always returns None, meaning a single scalar value.
        """
        return None

    def as_dist(self, max_weight=2.0, block_size=None):
        """Returns a :class:`pysnptools.distreader.DistReader` such that turns the an allele count of 0,1, or 2 into a probability distribution of
        [1,0,0],[0,1,0], or [0,0,1], respectively. Any other values produce a distribution of [NaN,NaN,NaN].

        :param max_weight: optional -- Tells the maximum allele count. Default of 2.
        :type max_weight: number

        :param block_size: optional -- Default of None (meaning to load all). Suggested number of sids to read into memory at a time.
        :type block_size: int or None

        :rtype: class:`DistReader`

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snpreader = Bed(bedfile, count_A1=False)
        >>> print(snpreader[0,0].read().val)
        [[2.]]
        >>> distreader = snpreader.as_dist(max_weight=2)
        >>> print(distreader[0,0].read().val)
        [[[0. 0. 1.]]]
        """
        from pysnptools.distreader._snp2dist import _Snp2Dist

        snp2dist = _Snp2Dist(self, max_weight=max_weight, block_size=block_size)
        return snp2dist


# LATER look at Eric's push related to Docker, etc from May
if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc t
    print("done")



from pysnptools.util import log_in_place
import numpy as np
import logging
import os
import unittest
import doctest
from pysnptools.snpreader import SnpReader


class SnpGen(SnpReader):
    """
    A :class:`.SnpReader` that generates deterministic "random" SNP data on the fly.

    See :class:`.SnpReader` for general examples of using SnpReader.

    **Constructor:**
        :Parameters: * **seed** (*number*) -- The random seed that (along with *block_size*) determines the SNP values.
        :Parameters: * **iid_count** (*number*) --  the number of iids (number of individuals)
        :Parameters: * **sid_count** (*number*) --  the number of sids (number of SNPs)
        :Parameters: * **chrom_count** (*number*) --  the number of chromosomes to generate (must be 22 or fewer)
        :Parameters: * **cache_file** (*string*) -- (default None) If provided, tells where to cache the common iid, sid, and pos information. Using it can save time.
        :Parameters: * **block_size** (*number* or None) -- (default None) Tells how many SNP to generate at once. The default picks a *block_size* such that *iid_count \* *block_size* is about 100,000.

        :Example:

        >>> from pysnptools.snpreader import SnpGen
        >>> #Prepare to generate data for 1000 individuals and 1,000,000 SNPs
        >>> snp_gen = SnpGen(seed=332,iid_count=1000,sid_count=1000*1000)
        >>> print(snp_gen.iid_count,snp_gen.sid_count)
        1000 1000000
        >>> snp_data = snp_gen[:,200*1000:201*1000].read() #Generate for all users and for SNPs 200K to 201K
        >>> print(snp_data.val[1,1], snp_data.iid_count, snp_data.sid_count)
        0.0 1000 1000


    :Also See: :func:`.snp_gen`

    """

    def __init__(
        self,
        seed,
        iid_count,
        sid_count,
        chrom_count=22,
        cache_file=None,
        block_size=None,
        sid_batch_size=None,
    ):
        self._ran_once = False
        self._cache_file = cache_file

        self._seed = seed
        self._iid_count = iid_count
        self._sid_count = sid_count
        self._chrom_count = chrom_count
        if sid_batch_size is not None:
            warnings.warn(
                "'sid_batch_size' is deprecated. Use 'block_size'", DeprecationWarning
            )
        self._block_size = (
            block_size or sid_batch_size or max((100 * 1000) // max(1, iid_count), 1)
        )

        if cache_file is not None:
            if not os.path.exists(cache_file):
                import pysnptools.util as pstutil

                pstutil.create_directory_if_necessary(cache_file)
                self._run_once()
                np.savez(
                    cache_file,
                    _row=self._row,
                    _col=self._col,
                    _col_property=self._col_property,
                )
            else:
                with np.load(cache_file, allow_pickle=True) as data:
                    self._row = data["_row"]
                    assert (
                        len(self._row) == iid_count
                    ), "The iid in the cache file has a different length than iid_count"
                    self._col = data["_col"]
                    assert (
                        len(self._col) == sid_count
                    ), "The sid in the cache file has a different length than sid_count"
                    self._col_property = data["_col_property"]
                    assert (
                        len(self._col_property) == sid_count
                    ), "The pos in the cache file has a different length than sid_count"
                    self._ran_once = True

    def __repr__(self):
        return "{0}(seed={1},iid_count={2},sid_count={3},chrom_count={4},block_size={5},cache_file={6})".format(
            self.__class__.__name__,
            self._seed,
            self._iid_count,
            self._sid_count,
            self._chrom_count,
            self._block_size,
            self._cache_file,
        )

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    _chrom_size = np.array(
        [
            263,
            255,
            214,
            203,
            194,
            183,
            171,
            155,
            145,
            144,
            144,
            143,
            114,
            109,
            106,
            98,
            92,
            85,
            67,
            72,
            50,
            56,
        ],
        dtype=np.int64,
    ) * int(
        1e6
    )  # The approximate size of human chromosomes in base pairs

    def _run_once(self):
        if self._ran_once:
            return
        self._ran_once = True
        self._row = np.array(
            [("0", "iid_{0}".format(i)) for i in range(self._iid_count)]
        )
        self._col = np.array(["sid_{0}".format(i) for i in range(self._sid_count)])
        self._col_property = np.full(((self._sid_count), 3), np.nan)

        chrom_total = SnpGen._chrom_size[: self._chrom_count].sum()
        step = chrom_total // self._sid_count
        start = 0
        chrom_size_so_far = 0
        for chrom_index in range(self._chrom_count):
            chrom_size_so_far += SnpGen._chrom_size[chrom_index]
            stop = chrom_size_so_far * self._sid_count // chrom_total
            self._col_property[start:stop, 0] = chrom_index + 1
            random_increment = np.random.RandomState(chrom_index ^ 2292).randint(step)
            self._col_property[start:stop, 2] = (
                np.arange(0, stop - start) * step + random_increment + 1
            )
            # print(chrom_index+1,start,stop,self._sid_count)
            start = stop

    def copyinputs(self, copier):
        self._run_once()
        copier.input(self._cache_file)

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,  # !!
        num_threads,
    ):
        self._run_once()
        import pysnptools.util as pstutil

        dtype = np.dtype(dtype)

        if order == "A":
            order = "F"

        row_index_count = (
            len(row_index_or_none) if row_index_or_none is not None else self._iid_count
        )  # turn to a count of the index positions e.g. all of them
        col_index = (
            col_index_or_none
            if col_index_or_none is not None
            else np.arange(self._sid_count)
        )  # turn to an array of index positions, e.g. 0,1,200,2200,10
        batch_index = (
            col_index // self._block_size
        )  # find the batch index of each index position, e.g. 0,0,0,2,0
        val = np.empty(
            (row_index_count, len(col_index)), order=order, dtype=dtype
        )  # allocate memory for result
        list_batch_index = list(set(batch_index))
        with log_in_place("working on snpgen batch", logging.INFO) as updater:
            for ii, i in enumerate(
                list_batch_index
            ):  # for each distinct batch index, generate snps
                updater("{0} of {1}".format(ii, len(list_batch_index)))
                start = int(i) * self._block_size  # e.g. 0 (then 2000)
                stop = start + self._block_size  # e.g. 1000, then 3000
                batch_val = self._get_val2(
                    start, stop, order=order, dtype=dtype
                )  # generate whole batch
                a = (
                    batch_index == i
                )  # e.g. [True,True,True,False,True], then [False,False,False,True,False]
                b = col_index[a] - start  # e.g.  0,1,200,10, then 200
                val[:, a] = (
                    batch_val[:, b]
                    if row_index_or_none is None
                    else pstutil.sub_matrix(batch_val, row_index_or_none, b)
                )

        return val

    def _get_dist(self):
        w = np.array([-0.6482249, -8.49790398])

        def dist_fit(x):
            log_y = w[0] * np.log(x) + w[1]
            return np.exp(log_y)

        x_sample = np.logspace(
            np.log10(0.1 / self._iid_count), np.log10(0.5), 100, base=10
        )  # discretize from 1e-7 (assuming iid_count=1e6) to .5, logarithmically
        y_sample = np.array(
            [dist_fit(x) for x in x_sample]
        )  # Find the relative weight of each point
        dist = y_sample / y_sample.sum()
        return x_sample, dist

    @staticmethod
    def _get_sid(sid_start, sid_stop):
        sid = ["sid_{0}".format(i) for i in range(sid_start, sid_stop)]
        return sid

    def _get_val2(self, sid_start, sid_stop, order, dtype):
        dtype = np.dtype(dtype)

        (
            x_sample,
            dist,
        ) = (
            self._get_dist()
        )  # The discrete distribution of minor allele frequencies (based on curve fitting to real data)
        val = SnpGen._get_val(
            x_sample,
            dist,
            self._iid_count,
            sid_start,
            sid_stop,
            self._seed,
            order,
            dtype,
        )
        return val

    @staticmethod
    def _get_val(x_sample, dist, iid_count, sid_start, sid_stop, seed, order, dtype):
        missing_rate = 0.218
        block_size_a = sid_stop - sid_start
        block_size_b = sid_stop - sid_start

        np.random.seed(seed + sid_start)
        val_a = np.empty((iid_count, block_size_a), order=order, dtype=dtype)
        sid_so_far = 0

        while sid_so_far < block_size_a:
            logging.debug(sid_so_far)
            sid_index_to_freq = np.random.choice(
                x_sample, size=block_size_b, replace=True, p=dist
            )  # For each sid, pick its minor allele freq
            val_b = (
                (np.random.rand(iid_count, 2, block_size_b) < sid_index_to_freq)
                .sum(axis=1)
                .astype(dtype)
            )  # Sample each allele and then sum the minor alleles # !!!slowest part
            missing = np.random.rand(iid_count, block_size_b) < missing_rate
            val_b[missing] = np.nan

            sid_index_to_minor_allele_count = np.nansum(
                val_b, axis=0
            )  # Find the # of minor alleles for each allele. By design this will be mostly 0's

            # Keep only the sids with a minor allele count between 1 and iid_count.
            legal = (sid_index_to_minor_allele_count != 0) * (
                sid_index_to_minor_allele_count <= iid_count
            )
            legal_count = legal.sum()
            val_a[:, sid_so_far : sid_so_far + legal_count] = val_b[:, legal][
                :, : block_size_a - sid_so_far
            ]  # !!! takes about 1/3 of the time. Faster if both were order="F"?
            sid_so_far += legal_count

        return val_a


class TestSnpGen(unittest.TestCase):
    np

    def test1(self):
        logging.info("in TestSnpGen test1")
        seed = 0
        snpgen = SnpGen(
            seed=seed, iid_count=1000, sid_count=1000 * 1000, block_size=1000
        )
        snpdata = snpgen[:, [0, 1, 200, 2200, 10]].read()
        snpdata2 = snpgen[:, [0, 1, 200, 2200, 10]].read()
        assert snpdata.allclose(snpdata2)

        from pysnptools.snpreader import Bed

        ref = Bed(
            os.path.dirname(os.path.realpath(__file__))
            + "/../../tests/datasets/snpgen.bed",
            count_A1=False,
        ).read()
        assert snpdata.allclose(ref, equal_nan=True)

        cache_file = "tempdir/cache_file_test1.npz"
        os.remove(cache_file) if os.path.exists(cache_file) else None
        snpgen3 = SnpGen(
            seed=seed,
            iid_count=1000,
            sid_count=1000 * 1000,
            block_size=1000,
            cache_file=cache_file,
        )
        snpdata3 = snpgen3[::10, [0, 1, 200, 2200, 10]].read()
        assert snpdata3.allclose(snpdata2[::10, :].read())
        snpgen4 = SnpGen(
            seed=seed,
            iid_count=1000,
            sid_count=1000 * 1000,
            block_size=1000,
            cache_file=cache_file,
        )
        snpdata4 = snpgen4[::10, [0, 1, 200, 2200, 10]].read()
        assert snpdata4.allclose(snpdata2[::10, :].read())


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSnpGen))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    result = doctest.testmod(optionflags=doctest.ELLIPSIS)
    assert result.failed == 0, "failed doc test: " + __file__



import os
import shutil
import logging
from contextlib import contextmanager
from pysnptools.util.filecache import FileCache

class LocalCache(FileCache):
    '''
    A :class:`.FileCache` for working with files locally.

    See :class:`.FileCache` for general examples of using FileCache.

    This is the simplest :class:`.FileCache` in that it stores everything on a local disk rather than storing things remotely.

    **Constructor:**
        :Parameters: * **directory** (*string*) -- The directory under which files should be written and read.

        :Example:

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> file_cache.save('sub1/file1.txt','Hello')
        >>> file_cache.file_exists('sub1/file1.txt')
        True

    '''
    def __init__(self,directory):
        super(LocalCache, self).__init__()
        self.directory =  os.path.normpath(directory).replace('\\','/')
        if os.path.exists(self.directory): assert not os.path.isfile(self.directory), "A directory cannot exist where a file already exists."

    def __repr__(self): 
        return "{0}('{1}')".format(self.__class__.__name__,self.directory)

    @property
    def name(self):
        '''
        A path-like name for this `LocalCache`.

        :rtype: string
        '''
        return self.directory

    def _simple_file_exists(self,simple_file_name):
        full_file = self.directory + "/" + simple_file_name
        return os.path.exists(full_file) and os.path.isfile(full_file)

    @contextmanager
    def _simple_open_read(self,simple_file_name,updater=None):
        logging.info("open_read('{0}')".format(simple_file_name))
        file_name = self.directory + "/" + simple_file_name
        assert os.path.exists(file_name), "File doesn't exist in LocalCache. File is '{0}'".format(file_name)

        yield file_name

        logging.info("close('{0}')".format(simple_file_name))

    @contextmanager
    def _simple_open_write(self,simple_file_name,size=0,updater=None):
        import pysnptools.util as pstutil # put here to avoid recursive nesting

        logging.info("open_write('{0}',size={1})".format(simple_file_name,size))

        #Register the file name in the directory
        file_name=self.directory + "/" + simple_file_name
        if os.path.exists(file_name): #This is only OK, if it is a directory containing no files (and thus doesn't exist)
            assert not os.path.isfile(file_name), "Can't open a file for write if it already exists."
            assert not self._at_least_one(self.walk(simple_file_name)), "Can't open a file for write if a directory with files already has the same name ({0},{1})".format(self,simple_file_name)
            shutil.rmtree(file_name)
        else:
            pstutil.create_directory_if_necessary(file_name,isfile=True)

        yield file_name

        logging.info("close('{0}')".format(simple_file_name))
        assert os.path.exists(file_name), "File doesn't exist in LocalCache. File is '{0}'".format(file_name)

    def _simple_rmtree(self,updater=None):
        logging.info("rmtree -- {0}".format(self.directory))
        if os.path.exists(self.directory):
            shutil.rmtree(self.directory)

    def _simple_remove(self,simple_file_name, updater=None):
        full_file = self.directory + "/" + simple_file_name
        assert os.path.exists(full_file) and os.path.isfile(full_file), "Expect file to exist (and be a file)"
        os.remove(full_file) #We don't actually remove the folders from the disk, but to this API the folder is gone.

    def _simple_getmtime(self,simple_file_name):
        full_file = self.directory + "/" + simple_file_name
        assert os.path.exists(full_file) and os.path.isfile(full_file), "Expect file to exist (and be a file)"
        return os.path.getmtime(full_file)

    def _simple_join(self,path):
        directory = self.directory + "/" + path
        if os.path.exists(directory): assert not os.path.isfile(directory), "Can't treat an existing file as a directory"
        return LocalCache(directory)

            
            

    def _simple_walk(self):
        for root, dirs, files in os.walk(self.directory):
            root = os.path.normpath(root).replace('\\','/')
            rel = os.path.relpath(root,self.directory).replace('\\','/')
            for file in files:
                yield file if rel == "." else rel + "/" + file



if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
import logging
import warnings
from pysnptools.standardizer import Standardizer

class BetaTrained(Standardizer):
    """A :class:`.Standardizer` to beta standardize one set of SNP data based on the mean of another set of SNP data

    See :class:`.Standardizer` for more information about standardization.

    **Constructor:**
        :Parameters: * **a** (*float*) -- The *a* parameter of the beta distribution
                     * **b** (*float*) -- The *b* parameter of the beta distribution
                     * **stats** (*ndarray of float*) -- The mean and stddev of each sid

    >>> from pysnptools.standardizer import Beta
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> train = Bed(bedfile,count_A1=False)[1:,:].read() # read SNP values for all but the first iid
    >>> _, betatrained = train.standardize(Beta(1,25),return_trained=True) #beta standardize and remember the mean and stddev of each sid
    >>> print(betatrained.stats[:5,:]) #Print the means and stddev of the first five sids
    [[...1.94983278  0.21828988]
     [...1.96989967  0.17086341]
     [...1.84280936  0.39057474]
     [...1.99665552  0.0577347 ]
     [...1.97658863  0.15120608]]
    >>> test = Bed(bedfile,count_A1=False)[0,:].read() # read SNP values for the first iid
    >>> test = test.standardize(betatrained) # Use the mean of the train data to beta standardize the test data.
    >>> print('{0:.6f}'.format(test.val[0,0]))
    0.681674
    """
    def __init__(self, a,b,sid,stats):
        super(BetaTrained, self).__init__()
        self.a=a
        self.b=b
        self.sid=sid
        self.stats=stats

    def __repr__(self): 
        return "{0}(a={1},b={2},stats={3},sid={4})".format(self.__class__.__name__,self.a,self.b,self.stats,self.sid)

    @property
    def is_constant(self):
        return True        

    def standardize(self, snps, block_size=None, return_trained=False, force_python_only=False, num_threads=None):
        if block_size is not None:
            warnings.warn("block_size is deprecated (and not needed, since standardization is in-place", DeprecationWarning)

        if hasattr(snps,"val"):
            val = snps.val
            assert np.array_equal(self.sid,snps.sid), "sid in training and use must be the same and in the same order"
        else:
            warnings.warn("standardizing an nparray instead of a SnpData is deprecated", DeprecationWarning) #LATER test coverage
            val = snps

        self._standardize_unit_and_beta(val, is_beta=True, a=self.a, b=self.b, apply_in_place=True,use_stats=True,stats=self.stats, num_threads=num_threads,
                                       force_python_only=force_python_only)
        if return_trained:
            return snps, self
        else:
            return snps

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE)



import numpy as np
import logging
from pysnptools.standardizer import Standardizer
from pysnptools.standardizer.unittrained import UnitTrained
import warnings
import pysnptools.util as pstutil

class Unit(Standardizer):
    """A :class:`.Standardizer` to unit standardize SNP data. For each sid, the mean of the values is zero with standard deviation 1.
    NaN values are then filled with zero, the mean (consequently, if there are NaN values, the final standard deviation will not be zero.

    See :class:`.Standardizer` for more information about standardization.

    >>> from pysnptools.standardizer import Unit
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snpdata1 = Bed(bedfile,count_A1=False).read().standardize(Unit())
    >>> print('{0:.6f}'.format(snpdata1.val[0,0]))
    0.229416
    """
    def __init__(self):
        super(Unit, self).__init__()

    def __repr__(self):
        return "{0}()".format(self.__class__.__name__)

    def standardize(self, snps, block_size=None, return_trained=False, force_python_only=False, num_threads=None):
        '''
        When cupy environment variable is set, will use cupy.
        '''
        xp = pstutil.array_module() # Get array module based on any ARRAY_MODULE environ variable.

        if block_size is not None:
            warnings.warn("block_size is deprecated (and not needed, since standardization is in-place", DeprecationWarning)

        if hasattr(snps,"val"):
            snps._val = xp.asarray(snps.val) #If cupy, replace SnpData's val with cupy array
            val = snps.val
        else:
            warnings.warn("standardizing an ndarray instead of a SnpData is deprecated", DeprecationWarning)
            val = snps

        stats = self._standardize_unit_and_beta(val, is_beta=False, a=np.nan, b=np.nan, apply_in_place=True,
                                                use_stats=False,stats=None,num_threads=num_threads, force_python_only=force_python_only)

        if return_trained:
            assert hasattr(snps,"val"), "return_trained=True requires that snps be a SnpData"
            return snps, UnitTrained(snps.sid, stats)
        else:
            return snps

    def _merge_trained(self, trained_list):
        sid = np.concatenate([trained.sid for trained in trained_list])
        stats = np.concatenate([trained.stats for trained in trained_list])
        return UnitTrained(sid, stats)

        
if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
import logging
import doctest
import unittest
import os.path
import time

from pysnptools.kernelreader import *
from pysnptools.snpreader import Bed
from pysnptools.util import create_directory_if_necessary
from pysnptools.pstreader import PstReader
from pysnptools.snpreader import SnpData
import pysnptools.standardizer as stdizer


class _fortesting_JustCheckExists(object):  # Implements ICopier
    def __init__(self, doPrintOutputNames=False):
        self.doPrintOutputNames = doPrintOutputNames

    def input(self, item):
        if isinstance(item, str):
            if not os.path.exists(item):
                raise Exception("Missing input file '{0}'".format(item))
        elif hasattr(item, "copyinputs"):
            item.copyinputs(self)
        # else -- do nothing

    def output(self, item):
        if isinstance(item, str):
            if not os.path.exists(item):
                raise Exception("Missing output file '{0}'".format(item))
            if self.doPrintOutputNames:
                print(item)
        elif hasattr(item, "copyoutputs"):
            item.copyoutputs(self)
        # else -- do nothing


class TestKernelReader(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))

    def test_merge_std(self):
        # unit vs beta
        for std in [stdizer.Beta(2, 10), stdizer.Unit()]:
            np.random.seed(0)
            sid_count = 20
            snpreader = SnpData(
                iid=[["0", "0"], ["1", "1"], ["2", "2"]],
                sid=[str(i) for i in range(sid_count)],
                val=np.array(
                    np.random.randint(3, size=[3, sid_count]),
                    dtype=np.float64,
                    order="F",
                ),
            )
            kerneldata0, trained0, diag0 = SnpKernel(
                snpreader, std, block_size=1
            )._read_with_standardizing(to_kerneldata=True, return_trained=True)
            kerneldata1, trained1, diag1 = SnpKernel(
                snpreader, std, block_size=None
            )._read_with_standardizing(to_kerneldata=True, return_trained=True)
            np.testing.assert_array_almost_equal(
                kerneldata0.val, kerneldata1.val, decimal=10
            )
            np.testing.assert_array_almost_equal(
                trained0.stats, trained1.stats, decimal=10
            )
            assert abs(diag0.factor - diag1.factor) < 1e-7

    def test_cpp_std(self):
        # Order C vs F
        for order in ["C", "F"]:
            # 32 vs 64
            for dtype in [np.float64, np.float32]:
                # unit vs beta
                for std in [stdizer.Unit(), stdizer.Beta(2, 10)]:
                    np.random.seed(0)
                    sid_count = 20
                    snpreader0 = SnpData(
                        iid=[["0", "0"], ["1", "1"], ["2", "2"]],
                        sid=[str(i) for i in range(sid_count)],
                        val=np.array(
                            np.random.randint(3, size=[3, sid_count]),
                            dtype=dtype,
                            order=order,
                        ),
                    )
                    snpreader1 = SnpData(
                        iid=[["3", "3"], ["4", "4"]],
                        sid=[str(i) for i in range(sid_count)],
                        val=np.array(
                            np.random.randint(3, size=[2, sid_count]),
                            dtype=dtype,
                            order=order,
                        ),
                    )

                    # has SNC
                    for has_SNC_in_train in [False, True]:
                        if has_SNC_in_train:
                            snpreader0.val[:, 1] = 0

                        # missing data
                        for has_missing_data in [False, True]:
                            if has_missing_data:
                                snpreader0.val[0, 2] = np.nan
                                snpreader1.val[0, 2] = np.nan

                            # gather stats vs not
                            cppa, stdcppa = snpreader0.read(
                                order=order, dtype=dtype
                            ).standardize(
                                std, return_trained=True, force_python_only=False
                            )
                            pya, stdpya = snpreader0.read(
                                order=order, dtype=dtype
                            ).standardize(
                                std, return_trained=True, force_python_only=True
                            )
                            np.testing.assert_array_almost_equal(
                                cppa.val,
                                pya.val,
                                decimal=10 if dtype == np.float64 else 5,
                            )

                            np.testing.assert_array_almost_equal(
                                stdcppa.stats,
                                stdpya.stats,
                                decimal=10 if dtype == np.float64 else 5,
                            )
                            assert (np.inf in stdcppa.stats[:, 1]) == has_SNC_in_train
                            assert (np.inf in stdpya.stats[:, 1]) == has_SNC_in_train

                            if has_SNC_in_train:
                                assert np.array_equal(
                                    cppa.val[:, 1], np.zeros([cppa.val.shape[0]])
                                )
                                assert np.array_equal(
                                    pya.val[:, 1], np.zeros([pya.val.shape[0]])
                                )

                            if has_missing_data:
                                assert 0 == cppa.val[0, 2]
                                assert 0 == pya.val[0, 2]

                            # uses stats
                            cppb = snpreader1.read(
                                order=order, dtype=dtype
                            ).standardize(stdcppa, force_python_only=False)
                            pyb = snpreader1.read(order=order, dtype=dtype).standardize(
                                stdpya, force_python_only=True
                            )
                            np.testing.assert_array_almost_equal(
                                cppb.val,
                                pyb.val,
                                decimal=10 if dtype == np.float64 else 5,
                            )
                            np.testing.assert_array_almost_equal(
                                stdcppa.stats,
                                stdpya.stats,
                                decimal=10 if dtype == np.float64 else 5,
                            )  # Make sure we haven't messed up the train stats

                            if has_SNC_in_train:
                                assert np.array_equal(
                                    cppb.val[:, 1], np.zeros([cppb.val.shape[0]])
                                )
                                assert np.array_equal(
                                    pyb.val[:, 1], np.zeros([pyb.val.shape[0]])
                                )

                            if has_missing_data:
                                assert cppb.val[0, 2] == 0
                                assert pyb.val[0, 2] == 0
        logging.info("done with 'test_cpp_std'")

    def test_intersection(self):
        from pysnptools.standardizer import Unit
        from pysnptools.kernelreader import SnpKernel
        from pysnptools.snpreader import Pheno
        from pysnptools.kernelreader._subset import _KernelSubset
        from pysnptools.snpreader._subset import _SnpSubset
        from pysnptools.util import intersect_apply

        snps_all = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        k = SnpKernel(snps_all, stdizer.Identity())

        pheno = Pheno(self.currentFolder + "/../examples/toydata.phe")
        pheno = pheno[1:, :]  # To test intersection we remove a iid from pheno

        k1, pheno = intersect_apply(
            [k, pheno]
        )  # SnpKernel is special because it standardizes AFTER intersecting.
        assert isinstance(k1.snpreader, _SnpSubset) and not isinstance(
            k1, _KernelSubset
        )

        # What happens with fancy selection?
        k2 = k[::2]
        assert isinstance(k2, SnpKernel)

        logging.info("Done with test_intersection")

    def test_respect_read_inputs(self):
        from pysnptools.kernelreader import KernelHdf5, Identity, KernelNpz, SnpKernel
        from pysnptools.standardizer import Unit
        from pysnptools.standardizer import Identity as StdIdentity
        from pysnptools.snpreader import Bed

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        iidref = KernelNpz("../examples/toydata.kernel.npz").iid

        for kernelreader in [
            SnpKernel(
                Bed("../examples/toydata.5chrom.bed", count_A1=True), StdIdentity()
            )[::2, ::2],
            Bed("../examples/toydata.5chrom.bed", count_A1=True)[::2, ::2].read_kernel(
                StdIdentity()
            ),
            KernelHdf5("../examples/toydata.kernel.hdf5"),
            Identity(iidref, test=[("0", "x"), ("0", "y")]),
            Identity(iidref),
            KernelNpz("../examples/toydata.kernel.npz"),
            KernelNpz("../examples/toydata.kernel.npz").read(),
            KernelNpz("../examples/toydata.kernel.npz")[::2, ::2],
            Bed("../examples/toydata.5chrom.bed", count_A1=True).read_kernel(Unit()),
            SnpKernel(Bed("../examples/toydata.5chrom.bed", count_A1=True), Unit()),
        ]:
            logging.info(str(kernelreader))
            for order in ["F", "C", "A"]:
                for dtype in [np.float32, np.float64]:
                    for force_python_only in [True, False]:
                        for view_ok in [True, False]:
                            val = kernelreader.read(
                                order=order,
                                dtype=dtype,
                                force_python_only=force_python_only,
                                view_ok=view_ok,
                            ).val
                            has_right_order = (
                                order == "A"
                                or (order == "C" and val.flags["C_CONTIGUOUS"])
                                or (order == "F" and val.flags["F_CONTIGUOUS"])
                            )
                            if hasattr(kernelreader, "val") and not view_ok:
                                assert kernelreader.val is not val
                            if (
                                hasattr(kernelreader, "val")
                                and view_ok
                                and kernelreader.val is not val
                                and (
                                    order == "A"
                                    or (
                                        order == "F"
                                        and kernelreader.val.flags["F_CONTIGUOUS"]
                                    )
                                    or (
                                        order == "C"
                                        and kernelreader.val.flags["C_CONTIGUOUS"]
                                    )
                                )
                                and (dtype is None or kernelreader.val.dtype == dtype)
                            ):
                                logging.info(
                                    "{0} could have read a view, but didn't".format(
                                        distreader
                                    )
                                )
                            assert val.dtype == dtype and has_right_order
        os.chdir(previous_wd)

    def test_respect_inputs(self):
        np.random.seed(0)
        for dtype_start, decimal_start in [(np.float32, 5), (np.float64, 10)]:
            for order_start in ["F", "C", "A"]:
                for sid_count in [20, 2]:
                    snpdataX = SnpData(
                        iid=[["0", "0"], ["1", "1"], ["2", "2"]],
                        sid=[str(i) for i in range(sid_count)],
                        val=np.array(
                            np.random.randint(3, size=[3, sid_count]),
                            dtype=dtype_start,
                            order=order_start,
                        ),
                    )
                    for stdx in [
                        stdizer.Beta(1, 25),
                        stdizer.Identity(),
                        stdizer.Unit(),
                    ]:
                        for snpreader0 in [snpdataX, snpdataX[:, 1:]]:
                            snpreader1 = snpreader0[1:, :]

                            (
                                refdata0,
                                trained_standardizer,
                            ) = snpreader0.read().standardize(
                                stdx, return_trained=True, force_python_only=True
                            )
                            refval0 = refdata0.val.dot(refdata0.val.T)
                            refdata1 = snpreader1.read().standardize(
                                trained_standardizer, force_python_only=True
                            )  # LATER why aren't these used?
                            refval1 = refdata0.val.dot(
                                refdata1.val.T
                            )  # LATER why aren't these used?
                            for dtype_goal, decimal_goal in [
                                (np.float32, 5),
                                (np.float64, 10),
                            ]:
                                for order_goal in ["F", "C", "A"]:
                                    k = snpreader0.read_kernel(
                                        standardizer=stdx,
                                        block_size=1,
                                        order=order_goal,
                                        dtype=dtype_goal,
                                    )
                                    PstReader._array_properties_are_ok(
                                        k.val, order_goal, dtype_goal
                                    )
                                    np.testing.assert_array_almost_equal(
                                        refval0,
                                        k.val,
                                        decimal=min(decimal_start, decimal_goal),
                                    )

    def test_fail(self):
        did_fail = True
        try:
            kd = KernelData(
                iid=[["0"], ["1"], ["2"]], val=[[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            )  # Wrong iid shape
            did_fail = False
        except Exception:
            pass
        assert (
            did_fail
        ), "The constructor should fail because the iid is the wrong shape"

    def test_kernel2(self):
        logging.info("in kernel2")
        kd = KernelData(
            iid=[["0", "0"], ["1", "1"], ["2", "2"]],
            val=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        )
        assert np.array_equal(
            kd.iid_to_index([["1", "1"], ["2", "2"]]), np.array([1, 2])
        )
        assert np.array_equal(
            kd.iid0_to_index([["1", "1"], ["2", "2"]]), np.array([1, 2])
        )
        kd = kd.standardize()
        assert np.abs(np.diag(kd.val).sum() - 3) < 1e-7
        assert kd.iid1_count == 3

    def test_snp_kernel2(self):
        logging.info("in test_snp_kernel2")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        snpkernel = SnpKernel(snpreader, standardizer=stdizer.Beta(1, 25))
        s = str(snpkernel)
        _fortesting_JustCheckExists().input(snpkernel)

    def test_npz(self):
        logging.info("in test_npz")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        kerneldata1 = snpreader.read_kernel(standardizer=stdizer.Unit())
        s = str(kerneldata1)
        output = "tempdir/kernelreader/toydata.kernel.npz"
        create_directory_if_necessary(output)
        KernelNpz.write(output, kerneldata1)
        kernelreader2 = KernelNpz(output)
        kerneldata2 = kernelreader2.read()
        np.testing.assert_array_almost_equal(
            kerneldata1.val, kerneldata2.val, decimal=10
        )
        logging.info("done with test")

    def test_subset(self):
        logging.info("in test_subset")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        snpkernel = SnpKernel(snpreader, stdizer.Unit())
        krsub = snpkernel[::2, ::2]
        kerneldata1 = krsub.read()
        expected = snpreader.read_kernel(stdizer.Unit())[::2].read()
        np.testing.assert_array_almost_equal(kerneldata1.val, expected.val, decimal=10)

        krsub2 = snpkernel[::2]
        kerneldata2 = krsub2.read()
        np.testing.assert_array_almost_equal(kerneldata2.val, expected.val, decimal=10)
        logging.info("done with test")

    def test_identity(self):
        logging.info("in test_identity")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        assert (
            np.array_equal(kid.row, kid.iid)
            and np.array_equal(kid.iid, kid.iid0)
            and np.array_equal(kid.iid0, kid.iid1)
            and np.array_equal(kid.iid1, kid.col)
        )
        np.testing.assert_array_almost_equal(
            kid.read().val, np.identity(snpreader.iid_count)
        )

        # subset
        sub1 = kid[1:5]
        np.testing.assert_array_almost_equal(sub1.read().val, np.identity(4))

        # zero size
        sub2 = kid[0:0]
        np.testing.assert_array_almost_equal(sub2.read().val, np.identity(0))

    def test_identity_sub(self):
        logging.info("in test_identity_sub")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        sub3 = kid[::2, 1:5]
        expected = np.identity(kid.iid_count)[::2, :][:, 1:5]
        np.testing.assert_array_almost_equal(sub3.read().val, expected)

        logging.info("done with test")

    def test_underscore_read1(self):
        logging.info("in test_underscore_read1")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        sub3 = kid[::2, ::2]
        expected = np.identity(kid.iid_count)[::2, :][:, ::2]
        np.testing.assert_array_almost_equal(sub3.read().val, expected)

        logging.info("done with test")

    def test_underscore_read2(self):
        logging.info("in test_underscore_read2")
        snpreader = Bed(
            self.currentFolder + "/../examples/toydata.5chrom.bed", count_A1=False
        )
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        sub3 = kid[::2, ::2]
        expected = np.identity(kid.iid_count)[::2, :][:, ::2]
        np.testing.assert_array_almost_equal(sub3.read().val, expected)

        logging.info("done with test")


# We do it this way instead of using doctest.DocTestSuite because doctest.DocTestSuite requires modules to be pickled, which python doesn't allow.
# We need tests to be pickleable so that they can be run on a cluster.
class TestKrDocStrings(unittest.TestCase):
    def test_kernelreader(self):
        import pysnptools.kernelreader.kernelreader

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.kernelreader.kernelreader, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpkernel(self):
        import pysnptools.kernelreader.snpkernel

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.kernelreader.snpkernel, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_kerneldata(self):
        import pysnptools.kernelreader.kerneldata

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.kernelreader.kerneldata, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_identity(self):
        import pysnptools.kernelreader.identity

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.kernelreader.identity,
            optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_kernelhdf5(self):
        import pysnptools.kernelreader.kernelhdf5

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.kernelreader.kernelhdf5, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_kernelnpz(self):
        import pysnptools.kernelreader.kernelnpz

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.kernelreader.kernelnpz, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestKernelReader))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestKrDocStrings))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()



import numpy as np
import logging
import warnings
from pysnptools.standardizer import Standardizer


class UnitTrained(Standardizer):
    """A :class:`.Standardizer` to unit standardize one set of SNP data based on the mean and stddev of another set of SNP data.
    NaN values are then filled with zero.

    See :class:`.Standardizer` for more information about standardization.

    **Constructor:**
        :Parameters: * **stats** (*ndarray of float*) -- The mean and stddev of each sid

    >>> from pysnptools.standardizer import Unit
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> train = Bed(bedfile,count_A1=False)[1:,:].read() # read SNP values for all but the first iid
    >>> _, unittrained = train.standardize(Unit(),return_trained=True) #Unit standardize and remember the mean and stddev of each sid
    >>> print(unittrained.stats[:5,:]) #Print the means and stddev of the first five sids
    [[...1.94983278  0.21828988]
     [...1.96989967  0.17086341]
     [...1.84280936  0.39057474]
     [...1.99665552  0.0577347 ]
     [...1.97658863  0.15120608]]
    >>> test = Bed(bedfile,count_A1=False)[0,:].read() # read SNP values for the first iid
    >>> test = test.standardize(unittrained) # Use the mean and stddev of the train data to unit standardize the test data.
    >>> print('{0:.6f}'.format(test.val[0,0]))
    0.229819
    """

    # !!might want to add code so that can check that sids are in the same order for both test and train
    def __init__(self, sid, stats):
        super(UnitTrained, self).__init__()
        self.sid = sid
        self.stats = stats
        self.sid_to_index = None

    def __repr__(self):
        return "{0}(stats={1},sid={2})".format(
            self.__class__.__name__, self.stats, self.sid
        )

    @property
    def is_constant(self):
        return True

    def standardize(
        self,
        snps,
        block_size=None,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        if block_size is not None:
            warnings.warn(
                "block_size is deprecated (and not needed, since standardization is in-place",
                DeprecationWarning,
            )

        if hasattr(snps, "val"):
            val = snps.val
            if len(self.sid) == len(snps.sid) and np.array_equal(self.sid, snps.sid):
                stats = self.stats
            else:
                if self.sid_to_index is None:
                    self.sid_to_index = {
                        sid: index for index, sid in enumerate(self.sid)
                    }
                stats = np.array(
                    [self.stats[self.sid_to_index[sid]] for sid in snps.sid]
                )
        else:
            warnings.warn(
                "standardizing an nparray instead of a SnpData is deprecated",
                DeprecationWarning,
            )  # LATER test coverage
            val = snps
            stats = self.stats

        self._standardize_unit_and_beta(
            val,
            is_beta=False,
            a=np.nan,
            b=np.nan,
            apply_in_place=True,
            use_stats=True,
            stats=stats,
            num_threads=num_threads,
            force_python_only=force_python_only,
        )

        if return_trained:
            return snps, self
        else:
            return snps


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.kernelreader import KernelReader
from pysnptools.pstreader import PstData
from pysnptools.pstreader import PstReader


class Identity(KernelReader):
    """
    A :class:`.KernelReader` for that represents an identity matrix. No memory for the values is allocated until :meth:`Identity.read` is called.

    See :class:`.KernelReader` for general examples of using KernelReaders.

    **Constructor:**
        :Parameters: * **iid** (an array of strings) -- The :attr:`KernelReader.iid` information

        :Example:

        >>> from pysnptools.kernelreader import Identity
        >>> identity = Identity(iid=[['fam0','iid0'],['fam0','iid1']])
        >>> print(identity.iid_count)
        2
        >>> print(identity.read().val) # '...' is possible space character
        [[...1.  0.]
         [...0.  1.]]

        >>> identity = Identity(iid=[['fam0','iid0'],['fam0','iid1'],['fam0','iid2']],test=[['fam0','iid1'],['fam0','iid3']])
        >>> print((identity.iid0_count, identity.iid1_count))
        (3, 2)
        >>> print(identity.read().val) # '...' is possible space character
        [[...0.  0.]
         [...1.  0.]
         [...0.  0.]]

    """

    def __init__(self, iid, test=None):  # !!! add docs and test for test
        super(Identity, self).__init__()

        if test is None:
            test = iid

        if test is iid:
            iid = PstData._fixup_input(
                iid,
                empty_creator=lambda ignore: np.empty([0, 2], dtype="str"),
                dtype="str",
            )
            test = iid
        else:
            iid = PstData._fixup_input(
                iid,
                empty_creator=lambda ignore: np.empty([0, 2], dtype="str"),
                dtype="str",
            )
            test = PstData._fixup_input(
                test,
                empty_creator=lambda ignore: np.empty([0, 2], dtype="str"),
                dtype="str",
            )

        if len(iid) > 0:
            self._row0 = iid
        else:
            self._row0 = self._empty

        if len(test) > 0:
            self._row1 = test
        else:
            self._row1 = self._empty

    _empty = np.empty([0, 2], dtype="str")

    def __repr__(self):
        return "{0}({1}x{2})".format(
            self.__class__.__name__, self.row_count, self.col_count
        )

    @property
    def row(self):
        return self._row0

    @property
    def col(self):
        return self._row1

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        dtype = np.dtype(dtype)
        if (
            row_index_or_none is None
            and col_index_or_none is None
            and self._row0 is self._row1
        ):  # read all of a square ID
            val = np.identity(self.row_count, dtype=dtype)
            if (order == "F" and not val.flags["F_CONTIGUOUS"]) or (
                order == "C" and not val.flags["C_CONTIGUOUS"]
            ):
                val = val.T
            return val
        else:  # Non-square
            # !!! This is also less efficient than it could be because it create a big identity matrix and then slices it.

            # In about O(col_count + row_count) fill in zeros
            big = np.zeros([self.row_count, self.col_count], dtype=dtype)
            common = set([PstReader._makekey(x) for x in self.row]) & set(
                [PstReader._makekey(x) for x in self.col]
            )
            big[self.row_to_index(common), self.col_to_index(common)] = 1.0
            val, shares_memory = self._apply_sparray_or_slice_to_val(
                big,
                row_index_or_none,
                col_index_or_none,
                order,
                dtype,
                force_python_only,
                num_threads,
            )
            return val

    def __getitem__(self, iid_indexer_and_snp_indexer):
        if isinstance(iid_indexer_and_snp_indexer, tuple):
            row_index_or_none, col_index_or_none = iid_indexer_and_snp_indexer
        else:
            row_index_or_none = iid_indexer_and_snp_indexer
            col_index_or_none = row_index_or_none

        # If iid0 will be iid1 then make 'is' equal
        iid = self.iid0[row_index_or_none]
        test = self.iid1[col_index_or_none]
        if len(iid) == len(test) and np.array_equal(iid, test):
            result = Identity(iid)
        else:
            result = Identity(iid=iid, test=test)
        return result


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



import numpy as np
import subprocess, sys, os.path
from itertools import *
import pandas as pd
import logging
from pysnptools.kernelreader import KernelReader
from pysnptools.pstreader import PstData
from pysnptools.kernelstandardizer import Identity as KS_Identity
from pysnptools.kernelstandardizer import DiagKtoN
import pysnptools.util as pstutil


class KernelData(KernelReader, PstData):
    """A :class:`.KernelReader` for holding kernel values in-memory, along with related iid information.
    It is usually created by calling the :meth:`.SnpReader.read_kernel` method on a :class:`.SnpReader`, for example, :class:`.Bed`.
    It can also be created by calling the :meth:`.KernelReader.read` method on :class:`.SnpKernel`, for example, :class:`.KernelNpz`.
    It can also be constructed directly.

    See :class:`.KernelReader` for details and examples.

    **Constructor:**
        :Parameters: * **iid** (an array of string pairs) -- The :attr:`KernelReader.iid` information.
                     * **iid0** (an array of string pairs) -- The :attr:`KernelReader.iid0` information.
                     * **iid1** (an array of strings pairs) -- The :attr:`KernelReader.iid1` information.
                     * **val** (a 2-D array of floats) -- The SNP values
                     * **name** (optional, string) -- Information to be display about the origin of this data
                     * **xp** (optional, a Python module or string) -- The array module that controls **val**,
                               for example, 'numpy' or 'cupy'. Defaults to numpy. Can also be set with the
                               'ARRAY_MODULE' environment variable.

        If *iid* is provided, don't provide *iid0* and *iid1*. Likewise, if *iid0* and *iid1* are provided, don't provide *iid*.

        :Example:

        >>> from pysnptools.kernelreader import KernelData
        >>> kerneldata = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,1.]])
        >>> print((kerneldata.val[0,1], kerneldata.iid_count))
        (0.5, 2)

    **Equality:**

        Two KernelData objects are equal if their three arrays (:attr:`.iid0`, :attr:`.iid1`, and :attr:`KernelData.val`) are 'array_equal'.
        (Their 'string' does not need to be the same).
        If either :attr:`.KernelData.val` contains NaN, the objects will not be equal. However, :meth:`.KernelData.allclose` can be used to treat NaN as
        regular values.


        :Example:

        >>> import numpy as np
        >>> from pysnptools.kernelreader import KernelData
        >>> kerneldata1 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,1.]])
        >>> kerneldata2 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,1.]])
        >>> print(kerneldata1 == kerneldata2) #True, because all the arrays have the same values.
        True
        >>> print(kerneldata1.val is kerneldata2.val) #False, because the two arrays have different memory.
        False
        >>> kerneldata3 = KernelData(iid=[['a','0'],['b','0']], val=[[1.,.5],[.5,1.]])
        >>> kerneldata4 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,1.]])
        >>> print(kerneldata3 == kerneldata4) #False, because the iids are different.
        False
        >>> kerneldata5 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,np.nan]])
        >>> kerneldata6 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,np.nan]])
        >>> print(kerneldata5 == kerneldata6) #False, because the val's contain NaN
        False
        >>> print(kerneldata5.allclose(kerneldata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True


    **Methods beyond** :class:`.KernelReader`
    """

    def __init__(
        self,
        iid=None,
        iid0=None,
        iid1=None,
        val=None,
        name=None,
        parent_string=None,
        xp=None,
    ):  # !!!autodoc doesn't generate good doc for this constructor
        # We don't have a 'super(KernelData, self).__init__()' here because KernelData takes full responsibility for initializing both its superclasses
        xp = pstutil.array_module(xp)
        self._val = None

        # !!why does SnpData __init__ have a copy_inputs, but KernelData doesn't?
        assert (iid is None) != (
            iid0 is None and iid1 is None
        ), "Either 'iid' or both 'iid0' 'iid1' must be provided."
        assert (
            name is None or parent_string is None
        ), "Can't set both 'name' and the deprecated 'parent_string'"
        if parent_string is not None:
            warnings.warn(
                "'parent_string' is deprecated. Use 'name'", DeprecationWarning
            )

        if iid is not None:
            self._row = PstData._fixup_input(
                iid,
                empty_creator=lambda ignore: np.empty([0, 2], dtype="str"),
                dtype="str",
            )
            self._col = self._row
        else:
            self._row = PstData._fixup_input(
                iid0,
                empty_creator=lambda ignore: np.empty([0, 2], dtype="str"),
                dtype="str",
            )
            self._col = PstData._fixup_input(
                iid1,
                empty_creator=lambda ignore: np.empty([0, 2], dtype="str"),
                dtype="str",
            )
        self._row_property = PstData._fixup_input(
            None,
            count=len(self._row),
            empty_creator=lambda count: np.empty([count, 0], dtype="str"),
            dtype="str",
        )
        self._col_property = PstData._fixup_input(
            None,
            count=len(self._col),
            empty_creator=lambda count: np.empty([count, 0], dtype="str"),
            dtype="str",
        )
        self._val = PstData._fixup_input_val(
            val,
            row_count=len(self._row),
            col_count=len(self._col),
            empty_creator=lambda row_count, col_count: xp.empty(
                [row_count, col_count], dtype=xp.float64
            ),
            xp=xp,
        )
        self._assert_iid0_iid1(check_val=True)
        self._name = name or parent_string or ""
        self._std_string_list = []
        self._xp = xp

    @property
    def val(self):
        """The 2D NumPy array of floats that represents the values of the kernel. You can get or set this property.

        >>> from pysnptools.kernelreader import KernelData
        >>> kerneldata = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,1.]])
        >>> print((kerneldata.val[0,1], kerneldata.iid_count))
        (0.5, 2)
        """
        return self._val

    @val.setter
    def val(self, new_value):
        self._val = PstData._fixup_input_val(
            new_value,
            row_count=len(self._row),
            col_count=len(self._col),
            empty_creator=lambda row_count, col_count: np.empty(
                [row_count, col_count], dtype=np.float64
            ),
            xp=self._xp,
        )
        self._assert_iid0_iid1(check_val=True)

    def allclose(self, value, equal_nan=True):
        """
        :param value: Other object with which to compare.
        :type value: :class:`KernelData`
        :param equal_nan: (Default: True) Tells if NaN in :attr:`KernelData.val` should be treated as regular values when testing equality.
        :type equal_nan: bool

        >>> import numpy as np
        >>> kerneldata5 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,np.nan]])
        >>> kerneldata6 = KernelData(iid=[['fam0','iid0'],['fam0','iid1']], val=[[1.,.5],[.5,np.nan]])
        >>> print(kerneldata5.allclose(kerneldata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True
        >>> print(kerneldata5.allclose(kerneldata6,equal_nan=False)) #False, if we consider the NaN as special values, all the arrays are not equal.
        False

        """
        return PstData.allclose(self, value, equal_nan=equal_nan)

    # !! SnpData.standardize() changes the str to help show that the data has been standardized. Should this to that too?
    def standardize(
        self,
        standardizer=DiagKtoN(),
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        """Does in-place standardization of the in-memory
        kernel data. The method multiples the values with a scalar factor such that the diagonal sums to iid_count. Although it works in place, for convenience
        it also returns the KernelData.

        :rtype: :class:`.KernelData` (standardizes in place, but for convenience, returns 'self')

        >>> from pysnptools.kernelreader import KernelNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> import numpy as np
        >>> kernel_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> kernel_on_disk = KernelNpz(kernel_file)
        >>> kerneldata1 = kernel_on_disk.read() # read all kernel values into memory
        >>> print(np.diag(kerneldata1.val).sum())
        5000000.0
        >>> kerneldata1.standardize() # standardize changes the values in kerneldata1.val
        KernelData(KernelNpz(...pysnptools/examples/toydata.kernel.npz'))
        >>> print(np.diag(kerneldata1.val).sum())
        500.0
        >>> kerneldata2 = kernel_on_disk.read().standardize() # Read and standardize in one expression with only one ndarray allocated.
        >>> print(np.diag(kerneldata2.val).sum())
        500.0
        """
        return standardizer.standardize(
            self,
            return_trained=return_trained,
            force_python_only=force_python_only,
            num_threads=num_threads,
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import os
import logging
import numpy as np
import sys
from contextlib import contextmanager
import time
import datetime
from types import ModuleType
import inspect
import warnings
from pysnptools.util.intrangeset import IntRangeSet
from bed_reader import get_num_threads, subset_f64_f64, subset_f32_f64, subset_f32_f32


def _testtest(data, iididx):
    return (data[0][iididx], data[1][iididx])


def intersect_apply(
    data_list, sort_by_dataset=True, intersect_before_standardize=True, is_test=False
):  # !!!add doc and tests for is_test
    """Intersects and sorts the iids from a list of datasets, returning new
    version of the datasets with all the same iids in the same order.

    :param data_list: list of datasets
    :type data_list: list
    :param sort_by_dataset: optional, If True (default), the iids are ordered
        according to the first non-None dataset.
        If False, the order is arbitrary, but consistent.
    :type sort_by_dataset: bool

    :param intersect_before_standardize: optional. Special code for
        :class:`.SnpKernel`, the class that postpones computing a kernel from SNP data.
        If True (default), :class:`.SnpKernel` will remove any iids before SNP
        standardization before computing the kernel.
        If False, SNPs will be standardized with all iids, then the kernel
        will be computed, then any iids will be removed.
    :type intersect_before_standardize: bool

    :rtype: list of datasets

    Here are the dataset formats understood and what is returned for each.

    ============================================== ================================================================
    Dataset Format                                 What is Returned
    ============================================== ================================================================
    None                                           None
    A :class:`.SnpReader`                          A new subsetting :class:`.SnpReader` with adjusted iid
    A :class:`.KernelReader`                       A new subsetting :class:`.KernelReader` with adjusted iid
    A dictionary with ['iid'] and ['vals'] keys    The same dictionary but with the iid and vals values adjusted
    Tuple of the form (val ndarray, iid list)      A new tuple with the val ndarray and iid list adjusted
    ============================================== ================================================================

    If the iids in all the datasets are already the same and in the same order, then the datasets are returned without change.

    Notice that only dictionaries are processed in-place. Inputting a :class:`.SnpReader` and :class:`.KernelReader` returns a new class of the same type (unless its iids
    are already ok). Inputting a tuple returns a new tuple (unless its iids are already ok).

    :Example:

    >>> from pysnptools.snpreader import Bed, Pheno
    >>> from pysnptools.kernelreader import SnpKernel
    >>> from pysnptools.standardizer import Unit
    >>> #Create five datasets in different formats
    >>> ignore_in = None
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> phenofile = example_file("tests/datasets/phenSynthFrom22.23.N300.randcidorder.txt")
    >>> covfile = example_file("tests/datasets/all_chr.maf0.001.covariates.N300.txt")
    >>> kernel_in = SnpKernel(Bed(bedfile,count_A1=False),Unit()) # Create a kernel from a Bed file
    >>> pheno_in = Pheno(phenofile,missing="")
    >>> cov = Pheno(covfile,missing="").read()
    >>> cov_as_tuple_in = (cov.val,cov.iid) #We could do cov directly, but as an example we make it a tuple.
    >>>
    >>> # Create five new datasets with consistent iids
    >>> ignore_out, kernel_out, pheno_out, cov_as_tuple_out = intersect_apply([ignore_in, kernel_in, pheno_in, cov_as_tuple_in])
    >>> # Print the first five iids from each dataset
    >>> print(ignore_out, kernel_out.iid[:5], pheno_out.iid[:5], cov_as_tuple_out[1][:5])
    None [['POP1' '0']
     ['POP1' '12']
     ['POP1' '44']
     ['POP1' '58']
     ['POP1' '65']] [['POP1' '0']
     ['POP1' '12']
     ['POP1' '44']
     ['POP1' '58']
     ['POP1' '65']] [['POP1' '0']
     ['POP1' '12']
     ['POP1' '44']
     ['POP1' '58']
     ['POP1' '65']]
    """

    iid_list = []
    reindex_list = []

    # LATER this doesn't cover non-square kernel readers. What should it do when there are two different iid lists?
    from pysnptools.kernelreader import SnpKernel
    from pysnptools.kernelreader import Identity as IdentityKernel

    for data in data_list:
        if data is None:
            iid = None
            reindex = lambda data, iididx: None
        elif intersect_before_standardize and isinstance(data, SnpKernel):
            iid = data.iid1 if is_test else data.iid0
            reindex = lambda data, iididx, is_test=is_test: _reindex_snpkernel(
                data, iididx, is_test
            )
        elif intersect_before_standardize and isinstance(data, IdentityKernel):
            iid = data.iid1 if is_test else data.iid0
            reindex = lambda data, iididx, is_test=is_test: _reindex_identitykernel(
                data, iididx, is_test
            )
        else:
            try:  # pheno dictionary
                iid = data["iid"]
                reindex = lambda data, iididx: _reindex_phen_dict(data, iididx)
            except Exception:
                if hasattr(data, "iid1") and is_test:  # test kernel
                    iid = data.iid1
                    reindex = lambda data, iididx: data[:, iididx]
                else:
                    try:
                        iid = data.iid
                        try:
                            if (
                                iid is data.col
                            ):  # If the 'iid' shares memory with the 'col', then it's a square-kernel-like thing and should be processed with just once index
                                reindex = lambda data, iididx: data[iididx]
                            else:
                                reindex = lambda data, iididx: data[iididx, :]
                        except Exception:
                            reindex = lambda data, iididx: data[iididx, :]
                    except AttributeError:  # tuple of (val,iid)
                        iid = data[1]
                        reindex = lambda data, iididx: _testtest(data, iididx)

        iid_list.append(iid)
        reindex_list.append(reindex)

    if len(iid_list) == 0:
        raise Exception("Expect a least one input item")

    if _all_same(iid_list):
        logging.debug("iids match up across {0} data sets".format(len(iid_list)))
        return data_list
    else:
        logging.debug("iids do not match up, so intersecting the data over individuals")
        indarr = intersect_ids(iid_list)
        assert (
            indarr.shape[0] > 0
        ), "no individuals remain after intersection, check that ids match in files"

        if sort_by_dataset:
            # Look for first non-None iid
            for i, iid in enumerate(iid_list):
                if iid is not None:
                    # sort the indexes so that SNPs ids in their original order (and
                    # therefore we have to move things around in memory the least amount)
                    sortind = np.argsort(indarr[:, i])
                    indarr = indarr[sortind]
                    break

        # !!! for the case in which some of the data items don't need to change, can we avoid calling reindex? Alternatively, should the _read code notice that all the iids are the same and in the same order
        data_out_list = []
        for i in range(indarr.shape[1]):
            data = data_list[i]
            iididx = indarr[:, i]
            reindex = reindex_list[i]
            new_data = reindex(data, iididx)
            data_out_list.append(new_data)

        return data_out_list


def _reindex_phen_dict(phen_dict, iididx):
    if len(phen_dict["vals"].shape) == 1:
        phen_dict["vals"] = phen_dict["vals"][iididx]
    else:
        phen_dict["vals"] = phen_dict["vals"][iididx, :]
    phen_dict["iid"] = phen_dict["iid"][iididx]
    return phen_dict


def _reindex_snpkernel(snpkernel, iididx, is_test=False):
    from pysnptools.kernelreader import SnpKernel

    if not is_test:
        new_reader = snpkernel.snpreader[iididx, :]
        result = SnpKernel(
            new_reader, snpkernel.standardizer, block_size=snpkernel.block_size
        )
    else:
        new_reader = snpkernel.test[iididx, :]
        result = SnpKernel(
            snpkernel.snpreader, snpkernel.standardizer, block_size=snpkernel.block_size
        )
    return result


def _reindex_identitykernel(identitykernel, iididx, is_test=False):
    from pysnptools.kernelreader import Identity as IdentityKernel

    if not is_test:
        iid = identitykernel.iid[iididx]
        result = IdentityKernel(iid)
    else:
        result = IdentityKernel(identitykernel.iid0, test=identitykernel.iid1[iididx])
    return result


def _all_same(iids_list):
    for i in range(len(iids_list) - 1):
        iidA = iids_list[i]
        iidB = iids_list[i + 1]
        if not np.array_equal(iidA, iidB):
            return False
    return True


def intersect_ids(idslist):
    """
    .. deprecated::
       Use :func:`.intersect_apply` instead.

    Takes a list of 2d string arrays of family and individual ids.
    These are intersected.

    :rtype: indarr, an array of size N x L, where N is the number of
        individuals in the intersection, and L is the number of lists in idslist, and which
        contains the index to use (in order) such that all people will be identical and in order
        across all data sets.

    If one of the lists=None, it is ignored (but still has values reported in indarr, all equal to -1),
    """
    id2ind = {}
    L = len(idslist)
    observed = np.zeros(L, dtype="bool")
    first = True
    for list_index, id_list in enumerate(idslist):
        if id_list is not None:
            observed[list_index] = 1
            if first:
                first = False
                for i in range(id_list.shape[0]):
                    id = (id_list[i, 0], id_list[i, 1])
                    entry = (
                        np.zeros(L) * np.nan
                    )  # id_list to contain the index for this id, for all lists provided
                    entry[list_index] = i  # index for the first one
                    id2ind[id] = entry
            else:
                for i in range(id_list.shape[0]):
                    id = (id_list[i, 0], id_list[i, 1])
                    if id in id2ind:
                        id2ind[id][list_index] = i

    indarr = np.array(
        list(id2ind.values()), dtype="float"
    )  # need float because may contain NaNs
    indarr[:, ~observed] = -1  # replace all Nan's from empty lists to -1
    inan = np.isnan(indarr).any(1)  # find any rows that contain at least one Nan
    indarr = indarr[~inan]  # keep only rows that are not NaN
    indarr = np.array(indarr, dtype="int")  # convert to int so can slice
    indarr = indarr[
        indarr[:, 0].argsort()
    ]  # make sure Py2 and Py3 return results in the same arbitrary order
    return indarr


def sub_matrix(
    val, row_index_list, col_index_list, order="A", dtype=np.float64, num_threads=None
):
    """
    Efficiently creates a sub-matrix from a 2-D or 3-D ndarray.

    :param val: The ndarray from which to copy.
    :type val: ndarray
    :param row_index_list: Tells which rows (and in which order) to copy into the sub-matrix
    :type row_index_list: list of integers
    :param col_index_list: Tells which columns (and in which order) to copy into the sub-matrix
    :type col_index_list: list of integers
    :param order: {'A' (default), 'C', 'F'}, optional -- Specify the order of the sub-matrix to create.
        If order is 'C', then the returned sub-matrix will be in C-contiguous order (first index varies the fastest).
        If order is 'F', then the array will be in F-contiguous order (second index varies the fastest).
        If order is 'A', then sub-matrix may be in any order F or C.
    :type order: string or None
    :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for sub-matrix created.
    :type dtype: data-type
    :param num_threads: optional -- The number of threads with which to work. Defaults to all available
        processors. Can also be set with these environment variables (listed in priority order):
        'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
    :type num_threads: None or int

    :rtype: ndarray

    >>> import numpy as np
    >>> import pysnptools.util as pstutil
    >>> np.random.seed(0) # set seed so that results are deterministic
    >>> matrix = np.random.rand(12,7) # create a 12 x 7 ndarray
    >>> submatrix = pstutil.sub_matrix(matrix,[0,2,11],[6,5,4,3,2,1,0])
    >>> print(int(submatrix.shape[0]),int(submatrix.shape[1]))
    3 7
    >>> print(matrix[2,0] == submatrix[1,6]) #The row # 2 is now #1, the column #0 is now #6.
    True

    Note: Behind the scenes, for performance, this function calls a Rust helper function.
    """

    if order == "A":
        if val.flags["F_CONTIGUOUS"]:
            effective_order = "F"
        else:
            effective_order = "C"
    else:
        effective_order = order
    dtype = np.dtype(dtype)

    original_dimensions = len(val.shape)
    if original_dimensions == 2:
        iid_count, sid_count = val.shape
        val_shape = 1
        val = val.reshape(iid_count, sid_count, 1)
    else:
        assert original_dimensions == 3, "Expect val dimensions of 2 or 3"
        iid_count, sid_count, val_shape = val.shape

    def create_sub_val(dtype):
        return np.full(
            (len(row_index_list), len(col_index_list), val_shape),
            np.NAN,
            dtype=dtype,
            order=effective_order,
        )

    logging.debug("About to call Rust matrixSubset")
    num_threads = get_num_threads(num_threads)
    if val.flags["F_CONTIGUOUS"] or val.flags["C_CONTIGUOUS"]:
        row_index_list = np.asarray(row_index_list, dtype=np.uintp)
        col_index_list = np.asarray(col_index_list, dtype=np.uintp)
        if val.dtype == np.float64:
            sub_val = create_sub_val(np.float64)
            subset_f64_f64(
                val,
                row_index_list,
                col_index_list,
                sub_val,
                num_threads,
            )
            if dtype == np.float64:
                pass
            elif dtype == np.float32:
                warnings.warn(
                    "Converting float64 to float32 can cause loss of information"
                )
                sub_val = sub_val.astype(dtype, order)
            else:
                raise Exception(
                    "dtype '{0}' not known, only float64 and float32".format(dtype)
                )
        elif val.dtype == np.float32:
            if dtype == np.float64:
                sub_val = create_sub_val(np.float64)
                subset_f32_f64(
                    val, row_index_list, col_index_list, sub_val, num_threads
                )
            elif dtype == np.float32:
                sub_val = create_sub_val(np.float32)
                subset_f32_f32(
                    val, row_index_list, col_index_list, sub_val, num_threads
                )
            else:
                raise Exception(
                    "dtype '{0}' not known, only float64 and float32".format(dtype)
                )
        else:
            raise Exception(
                "input dtype '{0}' not known, only float64 and float32".format(
                    val.dtype
                )
            )
    else:
        raise Exception("input order must be 'F' or 'C'")

    if original_dimensions == 2:
        assert sub_val.shape[2] == 1, "real assert"
        sub_val = sub_val.reshape(sub_val.shape[0], sub_val.shape[1])
    logging.debug("Back from Rust matrixSubset")
    return sub_val


def create_directory_if_necessary(name, isfile=True, robust=False):
    """
    Create a directory for a file if the directory doesn't already exist.

    :param name: file or directory name
    :type name: string
    :param isfile: If True (default), the name is a file, otherwise it is a directory.
    :type isfile: bool
    :param robust: If False (default), will try once to create the directory. If True, will try 25 times.
    :type robust: bool

    """
    if isfile:
        directory_name = os.path.dirname(name)
    else:
        directory_name = name

    if directory_name != "":
        if not robust:
            try:
                os.makedirs(directory_name)
            except OSError:
                if not os.path.isdir(directory_name):
                    raise Exception(
                        "not valid path: '{0}'. (Working directory is '{1}'".format(
                            directory_name, os.getcwd()
                        )
                    )
        else:
            is_ok = False
            time_to_sleep = 10.0
            for i in range(25):
                try:
                    os.makedirs(directory_name)
                    is_ok = True
                    break
                except OSError:
                    if not os.path.isdir(directory_name):
                        time_to_sleep *= 1.1
                        warnings.warning(
                            "creating directory robust=True, try#{0},time={3} error: not valid path: '{1}'. (Working directory is '{2}'".format(
                                i, directory_name, os.getcwd(), int(time_to_sleep)
                            )
                        )
                        time.sleep(int(time_to_sleep))
                        # make random?
                    else:
                        is_ok = True
                        break
            if not is_ok:
                raise Exception(
                    "not valid path: '{0}'. (Working directory is '{1}'".format(
                        directory_name, os.getcwd()
                    )
                )


def weighted_mean(ys, weights):
    """
    :param ys: The ndarray of values
    :type ys: ndarray
    :param weights: the weight of each value (unnormalized is fine)
    :type weights: ndarray
    :rtype: the weight mean


    >>> ys = np.array([103.664086,89.80645161,83.86888046,90.54141176])
    >>> weights = np.array([2.340862423,4.982888433,0.17522245,0.098562628])
    >>> round(weighted_mean(ys, weights),5)
    93.9487
    """
    mean = ys.dot(weights) / weights.sum()
    return mean


def weighted_simple_linear_regression(xs, ys, weights):
    """
    :param xs: The ndarray of independent values
    :type xs: ndarray
    :param ys: The ndarray of dependent values
    :type ys: ndarray
    :param weights: the weight of each case (unnormalized is fine)
    :type weights: ndarray
    :rtype: slope, intercept, xmean, ymean

    >>> xs = np.array([53.8329911,57.49486653,60.07392197,60.21081451])
    >>> ys = np.array([103.664086,89.80645161,83.86888046,90.54141176])
    >>> weights = np.array([2.340862423,4.982888433,0.17522245,0.098562628])
    >>> slope, intercept, xmean, ymean = weighted_simple_linear_regression(xs, ys, weights)
    >>> print(round(slope,5), round(intercept,5), round(xmean,5), round(ymean,5))
    -3.52643 293.05586 56.46133 93.9487

    """
    xmean = weighted_mean(xs, weights)
    xs_less_mean = xs - xmean
    ymean = weighted_mean(ys, weights)
    ys_less_mean = ys - ymean
    weighted_xs_less_mean = xs_less_mean * weights
    slope = ys_less_mean.dot(weighted_xs_less_mean) / xs_less_mean.dot(
        weighted_xs_less_mean
    )
    intercept = ymean - xmean * slope
    return slope, intercept, xmean, ymean


def print2(arg):
    """
    Make printing under Python3 look the same as under Python2.
    """
    s = str(arg).replace("b'", "'")
    print(s)


def format_delta(delta_seconds):
    """
    Format a time delta nicely.

    :param delta_seconds: The number of seconds
    :type delta_seconds: number
    :rtype: string

    >>> from pysnptools.util import format_delta
    >>> print(format_delta(86403.5))
    1 day, 0:00:03.500000
    """
    return datetime.timedelta(seconds=delta_seconds)


def _mbps(size, delta):
    return size * 8 / delta / 1e6


def _mbps_str(t0, size, total=0):
    delta = time.time() - t0
    mbps0 = _mbps(size, delta) if delta > 0 else 0
    percent = float(size) / total if total > 0 else 1
    left = delta / percent * (1 - percent) if total > 0 else 0
    return "{0:0.2f}, {1:0.1f}%, left={2}".format(
        mbps0, percent * 100, format_delta(left)
    )


@contextmanager
def log_in_place(name, level, time_lambda=time.time, show_log_diffs=False):
    """
    Create an one-argument function to write messages to. If the logging level
    is met, messages will appear. All messages will be on the same line and
    will include time.

    :Example:

    .. code-block:: python

        from pysnptools.util import log_in_place
        import logging
        import time
        logging.basicConfig(level=logging.INFO)

        with log_in_place("counting", logging.INFO) as updater:
        for i in range(100):
                updater(i)
                time.sleep(.1) #typically, some work -- not a delay -- goes here

    Outputs 100 messages on the same line, ending with
    something like "counting -- time=0:00:09.99, 99"
    """
    # !!! what if logging messages aren't suppose to go to stdout?
    t_wait = time_lambda()
    # We have to make this an array so that the value is by reference.
    last_len = [0]
    last_message_hash = [None]
    # Don't print the final newline if nothing is ever printed
    every_printed = [False]

    def writer(message):
        if logging.getLogger().level > level:
            return
        time_str = str(datetime.timedelta(seconds=time_lambda() - t_wait))
        if "." in time_str:
            time_str = time_str[
                : time_str.index(".") + 3
            ]  # Time to the 1/100th of a sec
        s = "{0} -- time={1}, {2}".format(name, time_str, message)
        if show_log_diffs:
            message_hash = hash(message)
            if (
                message_hash != last_message_hash[0]
                and last_message_hash[0] is not None
            ):
                sys.stdout.write("\n")
            last_message_hash[0] = message_hash
        # Pad with spaces to cover up previous message
        sys.stdout.write(f"{s}{' ' * max(0, last_len[0] - len(s))}\r")
        sys.stdout.flush()

        every_printed[0] = True
        last_len[0] = len(s)

    yield writer

    if logging.getLogger().level > level:
        return
    if every_printed[0]:
        sys.stdout.write("\n")


@contextmanager
def _file_transfer_reporter(name, size=None, updater=None):
    """
    If an update is given, we use that. Otherwise, we create our own.
    """
    if updater is None:
        bytes_so_far = [
            0
        ]  # We have to make this an array so that the value is by reference.
        t0 = time.time()

        with log_in_place(name, logging.INFO) as writer:

            def updater2(byte_size):
                bytes_so_far[0] += byte_size
                writer("Mbps={0}".format(_mbps_str(t0, bytes_so_far[0], size)))

            yield updater2
    else:
        yield updater


@contextmanager
def _multiopen(open_lambda, input_list):
    handle_list = [
        open_lambda(input) for input in input_list
    ]  # Open the related inputs
    list_to_yield = [
        handle.__enter__() for handle in handle_list
    ]  # Get the list to yield
    yield list_to_yield  # yield it
    for handle in handle_list:  # Close them
        handle.__exit__(None, None, None)


def _datestamp(appendrandom=False):
    import datetime

    now = datetime.datetime.now()
    s = str(now)[:19].replace(" ", "_").replace(":", "_")
    if appendrandom:
        import random

        s += "_" + str(random.random())[2:]
    return s


_warn_array_module_once = False


def array_module(xp=None):
    """
    Find the array module to use, for example **numpy** or **cupy**.

    :param xp: The array module to use, for example, 'numpy'
               (normal CPU-based module) or 'cupy' (GPU-based module).
               If not given, will try to read
               from the ARRAY_MODULE environment variable. If not given and
               ARRAY_MODULE is not set,
               will use numpy. If 'cupy' is requested, will
               try to 'import cupy'. If that import fails, will
               revert to numpy.
    :type xp: optional, string or Python module
    :rtype: Python module

    >>> from pysnptools.util import array_module
    >>> xp = array_module() # will look at environment variable
    >>> print(xp.zeros((3)))
    [0. 0. 0.]
    >>> xp = array_module('cupy') # will try to import 'cupy'
    >>> print(xp.zeros((3)))
    [0. 0. 0.]
    """
    xp = xp or os.environ.get("ARRAY_MODULE", "numpy")

    if isinstance(xp, ModuleType):
        return xp

    if xp == "numpy":
        return np

    if xp == "cupy":
        try:
            import cupy as cp

            return cp
        except ModuleNotFoundError as e:
            global _warn_array_module_once
            if not _warn_array_module_once:
                logging.warning(f"Using numpy. ({e})")
                _warn_array_module_once = True
            return np

    raise ValueError(f"Don't know ARRAY_MODULE '{xp}'")


def asnumpy(a):
    """
    Given an array created with any array module, return the equivalent
    numpy array. (Returns a numpy array unchanged.)

    >>> from pysnptools.util import asnumpy, array_module
    >>> xp = array_module('cupy')
    >>> zeros_xp = xp.zeros((3)) # will be cupy if available
    >>> zeros_np = asnumpy(zeros_xp) # will be numpy
    >>> zeros_np
    array([0., 0., 0.])
    """
    if isinstance(a, np.ndarray):
        return a
    return a.get()


def get_array_module(a):
    """
    Given an array, returns the array's
    module, for example, **numpy** or **cupy**.
    Works for numpy even when cupy is not available.

    >>> import numpy as np
    >>> zeros_np = np.zeros((3))
    >>> xp = get_array_module(zeros_np)
    >>> xp.ones((3))
    array([1., 1., 1.])
    """
    submodule = inspect.getmodule(type(a))
    module_name = submodule.__name__.split(".")[0]
    xp = array_module(module_name)
    return xp


from pysnptools.util.generate import snp_gen
from pysnptools.util._example_file import example_file

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    # There is also a unit test case in 'pysnptools\test.py'
    # that calls this doc test
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



class Runner:
    '''
    A Runner is an object that tells a :func:`.map_reduce` how to run on a local machine or cluster. Examples include
    :class:`.Local` and :class:`LocalMultiProc`.

    '''
    pass
